# Copyright 2019 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function, absolute_import
import six

import unittest
from nmosnode import registry
import time

# to run: python test_registry.py
# or, better: python -m unittest discover


# Because we don't have unittest.mock in 2.7, roll mock classes to cover only what we need.
class MockAggregator:

    def __init__(self):
        self.register_invocations = []
        self.unregister_invocations = []

    def register(self, *args, **kwargs):
        self.register_invocations.append([args, kwargs])

    def register_into(self, *args, **kwargs):
        self.register_invocations.append([args, kwargs])

    def unregister(self, *args, **kwargs):
        self.unregister_invocations.append([args, kwargs])

    def unregister_from(self, *args, **kwargs):
        self.unregister_invocations.append([args, kwargs])


class MockMDNSUpdater:

    def __init__(self):
        self.update_mdns_invocations = []

    def update_mdns(self, type, msg):
        self.update_mdns_invocations.append((type, msg))


class TestRegistryServices(unittest.TestCase):
    """Test registration and maintenance of services in the registry"""

    def setUp(self):
        """Runs before each test"""
        self.res_types = ["flow", "device"]
        self.node_data = {
            "label": "test_node", "href": 'http://127.0.0.1:88/', "host": "127.0.0.1", "services": [], "interfaces": []
        }
        self.registry = registry.FacadeRegistry(self.res_types, MockAggregator(),
                                                MockMDNSUpdater(), "test_node_id", self.node_data)

    def test_register_service(self):
        r = self.registry.register_service("test_srv", srv_type="test", pid=100, href="http://127.0.0.1:12345")
        self.assertEqual(registry.RES_SUCCESS, r)
        self.assertIn("test_srv", self.registry.list_services())
        self.assertEqual("http://127.0.0.1:12345", self.registry.get_service_href("test_srv"))
        service = self.registry.services["test_srv"]  # TODO: Does this expose too much guts?
        self.assertEqual(100, service["pid"])        # TODO: Need for "get_service_pid"?
        self.assertIn("resource", service)
        for rtype in self.res_types:
            self.assertEqual({}, service["resource"][rtype])

    def test_list_services(self):
        self.registry.register_service("test_srv_1", "test", 100, "http://127.0.0.1:12345")
        self.registry.register_service("test_srv_2", "test", 100, "http://127.0.0.1:12345")
        self.assertIn("test_srv_1", self.registry.list_services())
        self.assertIn("test_srv_2", self.registry.list_services())

    def test_get_service_href(self):
        self.registry.register_service("a", "test", 1, "a_href")
        self.assertEqual(registry.RES_NOEXISTS, self.registry.get_service_href("b"))
        self.assertEqual("a_href", self.registry.get_service_href("a"))

    def test_unregister_service(self):
        self.registry.register_service("test_srv_1", "test", 100, "http://127.0.0.1:12345")
        self.registry.register_service("test_srv_2", "test", 100, "http://127.0.0.1:12345")

        # do not allow non-existent services to be removed
        self.assertEqual(registry.RES_NOEXISTS, self.registry.unregister_service("not_there", 1))
        six.assertCountEqual(self, ["test_srv_1", "test_srv_2"], self.registry.list_services())

        # attempt to remove with wrong pid
        self.assertEqual(registry.RES_UNAUTHORISED, self.registry.unregister_service("test_srv_1", 1))
        six.assertCountEqual(self, ["test_srv_1", "test_srv_2"], self.registry.list_services())

        # remove one
        self.assertEqual(registry.RES_SUCCESS, self.registry.unregister_service("test_srv_1", 100))
        self.assertNotIn("test_srv_1", self.registry.services)
        six.assertCountEqual(self, ["test_srv_2"], self.registry.list_services())

    def test_update_service(self):
        self.registry.register_service("a", "test", 1, "http://a")
        self.assertEqual(registry.RES_NOEXISTS, self.registry.update_service("b", 1))
        self.assertEqual(["a"], self.registry.list_services())
        self.assertEqual(registry.RES_UNAUTHORISED, self.registry.update_service("a", 100, "blah"))
        self.assertEqual("http://a", self.registry.get_service_href("a"))
        self.assertEqual(registry.RES_SUCCESS, self.registry.update_service("a", 1, "updated"))
        self.assertAlmostEqual(time.time(), self.registry.services["a"]["heartbeat"], delta=0.04)
        self.assertEqual("updated", self.registry.get_service_href("a"))

    def test_heartbeat_service(self):
        self.registry.register_service("a", "test", 1, "http://a")
        self.assertEqual(registry.RES_NOEXISTS, self.registry.heartbeat_service("b", 1))
        self.assertEqual(registry.RES_UNAUTHORISED, self.registry.heartbeat_service("a", 100))
        self.assertEqual(registry.RES_SUCCESS, self.registry.heartbeat_service("a", 1))
        self.assertAlmostEqual(time.time(), self.registry.services["a"]["heartbeat"], delta=0.02)
        self.assertEqual("http://a", self.registry.get_service_href("a"))

    def test_cleanup_services(self):
        """Services with a heartbeat older than HEARTBEAT_TIMEOUT are removed"""
        self.registry.register_service("a", "test", 1, "a_href")
        self.registry.register_service("b", "test", 2, "b_href")
        self.registry.services["a"]["heartbeat"] = time.time() - registry.HEARTBEAT_TIMEOUT - 1
        self.registry.cleanup_services()
        self.assertEqual(["b"], self.registry.list_services())


class TestRegistry(unittest.TestCase):

    def setUp(self):
        """Runs before each test"""
        self.res_types = ["flow", "device", "sender"]
        self.mock_aggregator = MockAggregator()
        self.mock_mdns_updater = MockMDNSUpdater()
        self.node_data = {"label": "test", "href": "http://abcd", "host": "abcd", "services": [], "interfaces": []}
        self.registry = registry.FacadeRegistry(self.res_types, self.mock_aggregator,
                                                self.mock_mdns_updater, "test_node_id",
                                                self.node_data)

        # pre-populate with some services
        self.registry.register_service("a", srv_type="srv_a", pid=1)
        self.registry.register_service("b", srv_type="srv_b", pid=2)

        # ensure mock aggregator is clean
        self.mock_aggregator.register_invocations = []

    def test_register_device_adds_parent_facade(self):
        """Registering a resource adds a 'node_id' property to the resource"""
        self.registry.register_resource("a", 1, "device", "device_a_key", {"label": "device_a"})
        service_resources = self.registry.list_resource("device")
        self.assertEqual("test_node_id", service_resources["device_a_key"]["node_id"])

    def test_register_calls_aggregator(self):
        """When a resource is registered, the aggregator is informed"""
        self.registry.register_resource("a", 1, "flow", "flow_a_key", {"label": "flow_a"})
        expected_args = [('resource', 'flow', 'flow_a_key'), {'label': 'flow_a'}]
        self.assertEqual(self.mock_aggregator.register_invocations, [expected_args])

    def test_register_updates_mdns(self):
        """When a resource is registered, it is advertised vis mDNS"""
        self.registry.register_resource("a", 1, "flow", "flow_a_key", {"label": "flow_a"})
        expected_args = ('flow', 'register')
        self.assertEqual(self.mock_mdns_updater.update_mdns_invocations, [expected_args])

    def test_sender_manifest_returns_http(self):
        """Check that Sender manifest_href is not modified in HTTP mode"""
        registry.HTTPS_MODE = "disabled"
        self.registry.register_resource("a", 1, "sender", "sender_a_key", {"manifest_href": "http://some-url.com"})
        sender_resources = self.registry.list_resource("sender")
        scheme = sender_resources["sender_a_key"]["manifest_href"].split("://")[0]
        self.assertEqual(scheme, "http")

    def test_sender_manifest_returns_https(self):
        """Check that Sender manifest_href is modified in HTTPS mode"""
        registry.HTTPS_MODE = "enabled"
        self.registry.register_resource("a", 1, "sender", "sender_a_key", {"manifest_href": "http://some-url.com"})
        sender_resources = self.registry.list_resource("sender")
        scheme = sender_resources["sender_a_key"]["manifest_href"].split("://")[0]
        self.assertEqual(scheme, "https")

    def test_device_controls_return_http(self):
        """Check that Device control hrefs are unmodified in HTTP mode"""
        controls = [{"type": "some-type", "href": "http://some-url.com"},
                    {"type": "some-type", "href": "ws://some-url.com"}]
        registry.HTTPS_MODE = "disabled"
        self.registry.register_resource("a", 1, "device", "device_a_key", {"controls": controls,
                                                                           "max_api_version": "v1.2"})
        device_resources = self.registry.list_resource("device", "v1.2")
        self.assertEqual(len(controls), len(device_resources["device_a_key"]["controls"]))
        for control in device_resources["device_a_key"]["controls"]:
            self.assertIn(control["href"].split("://")[0], ["http", "ws"])

    def test_device_controls_return_https(self):
        """Check that Device control hrefs are modified in HTTPS mode"""
        controls = [{"type": "some-type", "href": "http://some-url.com"},
                    {"type": "some-type", "href": "ws://some-url.com"}]
        registry.HTTPS_MODE = "enabled"
        self.registry.register_resource("a", 1, "device", "device_a_key", {"controls": controls,
                                                                           "max_api_version": "v1.2"})
        device_resources = self.registry.list_resource("device", "v1.2")
        self.assertEqual(len(controls), len(device_resources["device_a_key"]["controls"]))
        for control in device_resources["device_a_key"]["controls"]:
            self.assertIn(control["href"].split("://")[0], ["https", "wss"])


if __name__ == '__main__':
    unittest.main()



# Copyright 2019 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from six import itervalues
from six import PY2

import unittest
import mock
import time

from nmosnode.aggregator import MDNSUpdater

MAX_ITERATIONS = 10


class TestMDNSUpdater(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super(TestMDNSUpdater, self).__init__(*args, **kwargs)
        if PY2:
            self.assertCountEqual = self.assertItemsEqual

    def setUp(self):
        self.mappings = {
            "device": "ver_dvc",
            "flow": "ver_flw",
            "source": "ver_src",
            "sender": "ver_snd",
            "receiver": "ver_rcv",
            "self": "ver_slf"}
        self.mdnstype = "_nmos-node._tcp"
        self.txt_recs = {"api_ver": "v1.0,v1.1,v1.2", "api_proto": "http"}
        self.mdnsname = "node_dummy_for_testing"
        self.port = 12345
        self.mdnsengine = mock.MagicMock()
        self.logger = mock.MagicMock()

        self.UUT = MDNSUpdater(
            self.mdnsengine,
            self.mdnstype,
            self.mdnsname,
            self.mappings,
            self.port,
            self.logger,
            txt_recs=self.txt_recs)

    def test_init(self):
        """Test of initialisation of an MDNSUpdater"""

        self.assertEqual(self.UUT.mdns_type, self.mdnstype)
        self.assertEqual(self.UUT.mdns_name, self.mdnsname)
        self.assertEqual(self.UUT.mappings, self.mappings)
        self.assertEqual(self.UUT.port, self.port)
        self.assertEqual(self.UUT.txt_rec_base, self.txt_recs)
        self.assertEqual(self.UUT.logger, self.logger)

        for key in itervalues(self.mappings):
            self.assertIn(key, self.UUT.service_versions)
            self.assertEqual(self.UUT.service_versions[key], 0)

        self.UUT.mdns.register.assert_called_once_with(self.mdnsname, self.mdnstype, self.port, self.txt_recs)

    def test_p2p_enable(self):
        """Test of MDNSUpdater.p2p_enable, should trigger an mdns update"""
        self.UUT.P2P_enable()

        self.assertTrue(self.UUT.p2p_enable)
        self.txt_recs.update(self.UUT.service_versions)
        counter = 0
        while not self.UUT._mdns_update_queue.empty() and counter < MAX_ITERATIONS:
            time.sleep(0.1)
            counter += 1
        self.UUT.mdns.update.assert_called_once_with(self.mdnsname, self.mdnstype, self.txt_recs)

    def test_inc_P2P_enable_count(self):
        """Test of of MDNSUpdater.inc_P2P_enable_count, should do nothing the first p2p_cut_in_count - 1 times,
        and then activate P2P mode"""

        for i in range(0, self.UUT.p2p_cut_in_count-1):
            self.UUT.inc_P2P_enable_count()

            self.assertFalse(self.UUT.p2p_enable)
            counter = 0
            while not self.UUT._mdns_update_queue.empty() and counter < MAX_ITERATIONS:
                time.sleep(0.1)
                counter += 1
            self.UUT.mdns.update.assert_not_called()

        self.UUT.inc_P2P_enable_count()
        self.assertTrue(self.UUT.p2p_enable)
        self.txt_recs.update(self.UUT.service_versions)
        counter = 0
        while not self.UUT._mdns_update_queue.empty() and counter < MAX_ITERATIONS:
            time.sleep(0.1)
            counter += 1
        self.UUT.mdns.update.assert_called_once_with(self.mdnsname, self.mdnstype, self.txt_recs)

    def test_P2P_disable_when_enabled(self):
        """When an MDNSUpdater is already enabled for P2P calling P2P_disable should disable P2P"""

        self.UUT.P2P_enable()
        counter = 0
        while not self.UUT._mdns_update_queue.empty() and counter < MAX_ITERATIONS:
            time.sleep(0.1)
            counter += 1
        self.UUT.mdns.update.reset_mock()
        self.UUT.P2P_disable()

        self.assertFalse(self.UUT.p2p_enable)
        counter = 0
        while not self.UUT._mdns_update_queue.empty() and counter < MAX_ITERATIONS:
            time.sleep(0.1)
            counter += 1
        self.UUT.mdns.update.assert_called_once_with(self.mdnsname, self.mdnstype, self.txt_recs)

    def test_P2P_disable_resets_enable_count(self):
        """If an MDNSUpdater is not yet enabled for P2P but has had inc_P2P_enable_count called on it then
        calling P2P_disable should reset this counter."""

        for i in range(0, self.UUT.p2p_cut_in_count-1):
            self.UUT.inc_P2P_enable_count()

            self.assertFalse(self.UUT.p2p_enable)
            counter = 0
            while not self.UUT._mdns_update_queue.empty() and counter < MAX_ITERATIONS:
                time.sleep(0.1)
                counter += 1
            self.UUT.mdns.update.assert_not_called()

        self.UUT.P2P_disable()

        for i in range(0, self.UUT.p2p_cut_in_count-1):
            self.UUT.inc_P2P_enable_count()

            self.assertFalse(self.UUT.p2p_enable)
            counter = 0
            while not self.UUT._mdns_update_queue.empty() and counter < MAX_ITERATIONS:
                time.sleep(0.1)
                counter += 1
            self.UUT.mdns.update.assert_not_called()

        self.UUT.inc_P2P_enable_count()
        self.assertTrue(self.UUT.p2p_enable)
        self.txt_recs.update(self.UUT.service_versions)
        counter = 0
        while not self.UUT._mdns_update_queue.empty() and counter < MAX_ITERATIONS:
            time.sleep(0.1)
            counter += 1
        self.UUT.mdns.update.assert_called_once_with(self.mdnsname, self.mdnstype, self.txt_recs)

    def test_update_mdns_does_nothing_when_not_enabled(self):
        """A call to MDNSUpdater.update_mdns should not do anything if P2P is disabled"""

        self.UUT.update_mdns("device", "register")
        self.assertEqual(self.UUT.service_versions[self.mappings["device"]], 0)
        counter = 0
        while not self.UUT._mdns_update_queue.empty() and counter < MAX_ITERATIONS:
            time.sleep(0.1)
            counter += 1
        self.UUT.mdns.update.assert_not_called()

    def test_update_mdns(self):
        """A call to MDNSUpdater.update_mdns when P2P is enabled ought to call mdns.update to increment version numbers
        for devices. Device version numbers should be 8-bit integers which roll over to 0 when incremented beyond
        the limits of 1 byte."""

        self.UUT.P2P_enable()

        # Don't check the mDNS update calls here as it'll take forever and it's handled by other tests already
        for i in range(1, 256):
            self.UUT.mdns.update.reset_mock()
            self.UUT.update_mdns("device", "register")
            self.assertEqual(self.UUT.service_versions[self.mappings["device"]], i)
            self.txt_recs.update(self.UUT.service_versions)

        self.UUT.mdns.update.reset_mock()
        while not self.UUT._mdns_update_queue.empty():
            self.UUT._mdns_update_queue.get()
        self.UUT.update_mdns("device", "register")
        self.assertEqual(self.UUT.service_versions[self.mappings["device"]], 0)
        self.txt_recs.update(self.UUT.service_versions)
        counter = 0
        while not self.UUT._mdns_update_queue.empty() and counter < MAX_ITERATIONS:
            time.sleep(0.1)
            counter += 1
        self.UUT.mdns.update.assert_called_once_with(self.mdnsname, self.mdnstype, self.txt_recs)



# Copyright 2019 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

from six import iteritems
from six.moves.urllib.parse import urljoin
from six import PY2

import unittest
import mock
import requests
import gevent
from copy import deepcopy
from nmosnode.aggregator import Aggregator, InvalidRequest, REGISTRATION_MDNSTYPE
from nmosnode.aggregator import AGGREGATOR_APINAMESPACE, LEGACY_REG_MDNSTYPE, AGGREGATOR_APINAME
from nmosnode.aggregator import ServerSideError
from nmosnode.aggregator import BACKOFF_INITIAL_TIMOUT_SECONDS, BACKOFF_MAX_TIMEOUT_SECONDS
from mdnsbridge.mdnsbridgeclient import NoService, EndOfServiceList
import nmosnode

MAX_ITERATIONS = 10


class TestAggregator(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super(TestAggregator, self).__init__(*args, **kwargs)
        if PY2:
            self.assertCountEqual = self.assertItemsEqual

    def setUp(self):
        paths = ['nmosnode.aggregator.Logger',
                 'nmosnode.aggregator.IppmDNSBridge',
                 'gevent.queue.Queue',
                 'gevent.spawn']
        patchers = {name: mock.patch(name) for name in paths}
        self.mocks = {name: patcher.start() for (name, patcher) in iteritems(patchers)}

        self.addCleanup(mock.patch.stopall)

        def printmsg(t):
            def _inner(msg):
                print(t + ": " + msg)
            return _inner

        # self.mocks['nmosnode.aggregator.Logger'].return_value.writeInfo.side_effect = printmsg("INFO")
        # self.mocks['nmosnode.aggregator.Logger'].return_value.writeWarning.side_effect = printmsg("WARNING")
        # self.mocks['nmosnode.aggregator.Logger'].return_value.writeDebug.side_effect = printmsg("DEBUG")
        # self.mocks['nmosnode.aggregator.Logger'].return_value.writeError.side_effect = printmsg("ERROR")
        # self.mocks['nmosnode.aggregator.Logger'].return_value.writeFatal.side_effect = printmsg("FATAL")

    def construct_url(self, aggregator, api_ver, path):
        """Constructs URL from aggregator and path, along with API version and namespace"""
        return "{}/{}/{}/{}/{}".format(aggregator, AGGREGATOR_APINAMESPACE, AGGREGATOR_APINAME, api_ver, path)

    def test_init(self):
        """Test a call to Aggregator()"""
        self.mocks['gevent.spawn'].side_effect = lambda f: mock.MagicMock(thread_function=f)

        a = Aggregator()

        self.mocks['nmosnode.aggregator.Logger'].assert_called_once_with('aggregator_proxy', None)
        self.assertEqual(a.logger, self.mocks['nmosnode.aggregator.Logger'].return_value)
        self.mocks['nmosnode.aggregator.IppmDNSBridge'].assert_called_once_with(logger=a.logger)
        self.assertEqual(a._reg_queue, self.mocks['gevent.queue.Queue'].return_value)
        self.assertEqual(a.main_thread.thread_function, a._main_thread)
        self.assertEqual(a.queue_thread.thread_function, a._process_queue)

    def test_set_api_version(self):
        """Test that setting the api version also sets the correct DNS-SD service type"""
        a = Aggregator()
        versions_service_type = [
            {'api_ver': 'v1.0', 'srv_type': LEGACY_REG_MDNSTYPE},
            {'api_ver': 'v1.1', 'srv_type': LEGACY_REG_MDNSTYPE},
            {'api_ver': 'v1.2', 'srv_type': LEGACY_REG_MDNSTYPE},
            {'api_ver': 'v1.3', 'srv_type': REGISTRATION_MDNSTYPE},
        ]

        for x in versions_service_type:
            a._set_api_version_and_srv_type(x['api_ver'])

            self.assertEqual(a.aggregator_apiversion, x['api_ver'])
            self.assertEqual(a.service_type, x['srv_type'])

    def test_register_into(self):
        """register_into() should register an object into a namespace, adding a scheduled call to the registration
        queue to that effect."""
        a = Aggregator()

        namespace = "potato"
        objects = [("dummy", "testkey", {"test_param": "test_value", "test_param1": "test_value1"})]

        for o in objects:
            a.register_into(namespace, o[0], o[1], **o[2])
            a._reg_queue.put.assert_called_with({
                "method": "POST",
                "namespace": namespace,
                "res_type": o[0],
                "key": o[1]})
            send_obj = {"type": o[0], "data": {k: v for (k, v) in iteritems(o[2])}}
            if 'id' not in send_obj['data']:
                send_obj['data']['id'] = o[1]
            self.assertEqual(a._node_data["entities"][namespace][o[0]][o[1]], send_obj)

    def test_register(self):
        """register() should register an object into namespace "resource", adding a scheduled call to the registration
        queue to that effect. There is special behaviour when registering a node, since the object can only ever have
        one node registration at a time."""

        a = Aggregator()

        namespace = "resource"
        objects = [
            ("dummy", "testkey", {"test_param": "test_value", "test_param1": "test_value1"}),
            ("node", "testnode", {"test_param": "test_value", "test_param1": "test_value1"})
        ]
        with mock.patch.object(nmosnode.aggregator, 'OAUTH_MODE', return_value=True):
            for o in objects:
                a.register(o[0], o[1], **o[2])

                send_obj = {"type": o[0], "data": {k: v for (k, v) in iteritems(o[2])}}
                if 'id' not in send_obj['data']:
                    send_obj['data']['id'] = o[1]
                if o[0] == "node":
                    self.assertEqual(a._node_data["node"], send_obj)
                else:
                    a._reg_queue.put.assert_called_with({
                        "method": "POST",
                        "namespace": namespace,
                        "res_type": o[0],
                        "key": o[1]})

                    self.assertEqual(a._node_data["entities"][namespace][o[0]][o[1]], send_obj)

    def test_unregister(self):
        """unregister() should schedule a call to unregister the specified devices.
        Special behaviour is expected when unregistering a node, where the request should be made immediately."""
        a = Aggregator()

        namespace = "resource"
        objects = [
            ("dummy", "testkey", {"test_param": "test_value", "test_param1": "test_value1"}),
            ("node", "testnode", {"test_param": "test_value", "test_param1": "test_value1"})
        ]
        with mock.patch.object(nmosnode.aggregator, 'OAUTH_MODE', return_value=True):
            for o in objects:
                with mock.patch.object(a, '_unregister_node') as un_register:
                    a.register(o[0], o[1], **o[2])

                    a.unregister(o[0], o[1])

                    if o[0] == "node":
                        self.assertIsNone(a._node_data["node"])
                    else:
                        un_register.assert_not_called()
                        a._reg_queue.put.assert_called_with({
                            "method": "DELETE",
                            "namespace": namespace,
                            "res_type": o[0],
                            "key": o[1]})
                        self.assertNotIn(o[1], a._node_data["entities"][namespace][o[0]])

    def test_stop(self):
        """A call to stop should set _running to false and then join the heartbeat thread."""
        self.mocks['gevent.spawn'].side_effect = lambda f: mock.MagicMock(thread_function=f)
        a = Aggregator()

        self.assertTrue(a._running)

        a.stop()

        self.assertFalse(a._running)
        a.main_thread.join.assert_called_with()
        a.queue_thread.join.assert_called_with()

    def test_get_aggregator_returns_services_in_order(self):
        """Test that each aggregator is returned in order"""
        a = Aggregator(mdns_updater=mock.MagicMock())
        test_aggregators = ['http://example0.com/aggregator/',
                            'http://example1.com/aggregator/',
                            'http://example2.com/aggregator/',
                            'http://example3.com/aggregator/']

        a.mdnsbridge.getHrefWithException.side_effect = test_aggregators

        for agg in test_aggregators:
            self.assertEqual(a._get_aggregator(), agg)

    def test_get_aggregator_returns_none_when_end_of_list(self):
        """Test that None is returned and backoff period is increased when EndofServiceList Exception raised"""
        a = Aggregator(mdns_updater=mock.MagicMock())

        a.mdnsbridge.getHrefWithException.side_effect = EndOfServiceList()

        self.assertEqual(a._get_aggregator(), None)
        self.assertEqual(a._backoff_period, BACKOFF_INITIAL_TIMOUT_SECONDS)

    def test_get_aggregator_uses_correct_srv_type(self):
        """Test that _get_aggregator uses the correct service type for API version and protocol type"""
        a = Aggregator(mdns_updater=mock.MagicMock())

        aggregator = "http://example0.com/aggregator/"
        versions = ['v1.0', 'v1.1', 'v1.2', 'v1.3']
        expected_calls = [
            mock.call(LEGACY_REG_MDNSTYPE, None, versions[0], 'http', False),
            mock.call(LEGACY_REG_MDNSTYPE, None, versions[1], 'http', False),
            mock.call(LEGACY_REG_MDNSTYPE, None, versions[2], 'http', False),
            mock.call(REGISTRATION_MDNSTYPE, None, versions[3], 'http', False),
            mock.call(LEGACY_REG_MDNSTYPE, None, versions[0], 'https', False),
            mock.call(LEGACY_REG_MDNSTYPE, None, versions[1], 'https', False),
            mock.call(LEGACY_REG_MDNSTYPE, None, versions[2], 'https', False),
            mock.call(REGISTRATION_MDNSTYPE, None, versions[3], 'https', False),
        ]
        a.mdnsbridge.getHrefWithException.return_value = aggregator
        nmosnode.aggregator.OAUTH_MODE = False

        for v in versions:
            a._set_api_version_and_srv_type(v)
            self.assertEqual(a._get_aggregator(), aggregator)

        with mock.patch.dict(nmosnode.aggregator._config, {'https_mode': "enabled"}):
            for v in versions:
                a._set_api_version_and_srv_type(v)
                self.assertEqual(a._get_aggregator(), aggregator)

        a.mdnsbridge.getHrefWithException.assert_has_calls(expected_calls)

    def test_get_aggregator_returns_none_when_no_aggregators(self):
        """Test that None is returned, increment P2P counter and backoff period is increased
        when NoService Exception raised"""
        a = Aggregator(mdns_updater=mock.MagicMock())

        a.mdnsbridge.getHrefWithException.side_effect = NoService()

        self.assertEqual(a._get_aggregator(), None)
        self.assertEqual(a._backoff_period, BACKOFF_INITIAL_TIMOUT_SECONDS)
        a._mdns_updater.inc_P2P_enable_count.assert_called_with()

    def test_reset_backoff(self):
        """Test backoff period is reset to zero"""
        a = Aggregator(mdns_updater=mock.MagicMock())
        a._backoff_period = 100

        a._reset_backoff_period()
        self.assertEqual(a._backoff_period, 0)

    def test_backoff_thread_timout(self):
        """Check that the backoff thread, sets and un-sets the backoff flag after backoff period"""
        a = Aggregator(mdns_updater=mock.MagicMock())

        with mock.patch('gevent.sleep') as timer:
            a._backoff_period = BACKOFF_INITIAL_TIMOUT_SECONDS
            a._back_off_timer()
            timer.assert_called_with(BACKOFF_INITIAL_TIMOUT_SECONDS)
            self.assertFalse(a._backoff_active)

    def test_increase_backoff_period(self):
        """Check that the back off period is increased exponentially from its minimum to its maximum"""
        a = Aggregator(mdns_updater=mock.MagicMock())

        expected_values = []
        for x in range(0, 3):
            expected_values.append((BACKOFF_INITIAL_TIMOUT_SECONDS) * 2**x)
        expected_values.append(BACKOFF_MAX_TIMEOUT_SECONDS)
        expected_values.append(BACKOFF_MAX_TIMEOUT_SECONDS)

        for x in expected_values:
            a._aggregator_list_stale = False
            a._increase_backoff_period()
            self.assertEqual(a._backoff_period, x)
            self.assertEqual(a._aggregator_list_stale, True)

    # # ================================================================================================================
    # # Test heartbeat operation
    # # ================================================================================================================

    def test_heartbeat_200_when_registered(self):
        """Test heartbeat operation when heartbeat request returns HTTP 200 when registered
        After HTTP 200 heartbeat, Node should still be registered and should wait 5 seconds"""
        DUMMYNODEID = "90f7c2c0-cfa9-11e7-9b9d-2fe338e1e7ce"
        a = Aggregator(mdns_updater=mock.MagicMock())
        a._node_data["registered"] = True
        a._node_data["node"] = {"type": "node", "data": {"id": DUMMYNODEID}}
        a.aggregator = "http://example.com"

        def killloop(*args, **kwargs):
            a._running = False

        def request(*args, **kwargs):
            return mock.MagicMock(status_code=200)

        with mock.patch('gevent.sleep', side_effect=killloop) as sleep:
            with mock.patch.object(a, '_send', side_effect=request) as send:
                return_val = a._heartbeat()

                self.assertTrue(return_val)
                send.assert_called_with("POST", "http://example.com", a.aggregator_apiversion,
                                        "health/nodes/{}".format(DUMMYNODEID))
                a._mdns_updater.inc_P2P_enable_count.assert_not_called()
                self.assertTrue(a._node_data["registered"])
                self.assertTrue(a._aggregator_list_stale)
                self.assertEqual(a._backoff_period, 0)
                sleep.assert_called_with(1)

    def test_heartbeat_200_when_not_registered(self):
        """Test heartbeat operation when heartbeat request returns HTTP 200 when not registered
        After HTTP 200 heartbeat, Node should unregister Node then perform re-registration with same aggregator"""
        DUMMYNODEID = "90f7c2c0-cfa9-11e7-9b9d-2fe338e1e7ce"
        a = Aggregator(mdns_updater=mock.MagicMock())
        a._node_data["registered"] = False
        a._node_data["node"] = {"type": "node", "data": {"id": DUMMYNODEID}}
        a.aggregator = "http://example.com"

        def request(*args, **kwargs):
            return mock.MagicMock(status_code=200, headers={'Location': 'path/xxx'})

        with mock.patch.object(a, '_send', side_effect=request) as send:
            with mock.patch.object(a, '_unregister_node', return_value=True) as un_reg:
                with mock.patch.object(a, '_register_node', return_value=True) as register:
                    return_val = a._heartbeat()

                    self.assertTrue(return_val)
                    send.assert_called_with("POST", "http://example.com", a.aggregator_apiversion,
                                            "health/nodes/{}".format(DUMMYNODEID))
                    a._mdns_updater.inc_P2P_enable_count.assert_not_called()
                    un_reg.assert_called_once_with('path/xxx')
                    register.assert_called_once_with(a._node_data["node"])

    def test_heartbeat_200_when_not_registered_failure(self):
        """Test heartbeat operation when heartbeat request returns HTTP 200 when not registered
        and aggregator in failed state
        After HTTP 200 heartbeat, Node should attempt to unregister node then return"""
        DUMMYNODEID = "90f7c2c0-cfa9-11e7-9b9d-2fe338e1e7ce"
        a = Aggregator(mdns_updater=mock.MagicMock())
        a._node_data["registered"] = False
        a._node_data["node"] = {"type": "node", "data": {"id": DUMMYNODEID}}
        a.aggregator = "http://example.com"

        def request(*args, **kwargs):
            return mock.MagicMock(status_code=200, headers={'Location': 'path/xxx'})

        with mock.patch.object(a, '_send', side_effect=request) as send:
            with mock.patch.object(a, '_unregister_node', return_value=False) as un_reg:
                return_val = a._heartbeat()

                self.assertFalse(return_val)
                send.assert_called_with("POST", "http://example.com", a.aggregator_apiversion,
                                        "health/nodes/{}".format(DUMMYNODEID))
                a._mdns_updater.inc_P2P_enable_count.assert_not_called()
                un_reg.assert_called_once_with('path/xxx')

    def test_heartbeat_409(self):
        """Test heartbeat operation when heartbeat request returns HTTP 409
        After HTTP 409 heartbeat, Node should unregister Node then perform re-registration with same aggregator"""
        DUMMYNODEID = "90f7c2c0-cfa9-11e7-9b9d-2fe338e1e7ce"
        a = Aggregator(mdns_updater=mock.MagicMock())
        a._node_data["registered"] = True
        a._node_data["node"] = {"type": "node", "data": {"id": DUMMYNODEID}}
        a.aggregator = "http://example.com"

        def request(*args, **kwargs):
            return mock.MagicMock(status_code=409, headers={'Location': 'path/xxx'})

        with mock.patch.object(a, '_send', side_effect=request) as send:
            with mock.patch.object(a, '_unregister_node', return_value=True) as un_reg:
                with mock.patch.object(a, '_register_node', return_value=True) as register:
                    return_val = a._heartbeat()

                    self.assertTrue(return_val)
                    send.assert_called_with("POST", "http://example.com", a.aggregator_apiversion,
                                            "health/nodes/{}".format(DUMMYNODEID))
                    a._mdns_updater.inc_P2P_enable_count.assert_not_called()
                    un_reg.assert_called_once_with('path/xxx')
                    register.assert_called_once_with(a._node_data["node"])

    def test_heartbeat_409_failure(self):
        """Test heartbeat operation when heartbeat request returns HTTP 409 and aggregator in failed state
        After HTTP 409 heartbeat, Node should attempt to unregister Node and return"""
        DUMMYNODEID = "90f7c2c0-cfa9-11e7-9b9d-2fe338e1e7ce"
        a = Aggregator(mdns_updater=mock.MagicMock())
        a._node_data["registered"] = True
        a._node_data["node"] = {"type": "node", "data": {"id": DUMMYNODEID}}
        a.aggregator = "http://example.com"

        def request(*args, **kwargs):
            return mock.MagicMock(status_code=409, headers={'Location': 'path/xxx'})

        with mock.patch.object(a, '_send', side_effect=request) as send:
            with mock.patch.object(a, '_unregister_node', return_value=False) as un_reg:
                return_val = a._heartbeat()

                self.assertFalse(return_val)
                send.assert_called_with("POST", "http://example.com", a.aggregator_apiversion,
                                        "health/nodes/{}".format(DUMMYNODEID))
                a._mdns_updater.inc_P2P_enable_count.assert_not_called()
                un_reg.assert_called_once_with('path/xxx')

    def test_heartbeat_404(self):
        """Test heartbeat operation when heartbeat request returns HTTP 404
        After HTTP 404, Node should mark itself as not registered and attempt re-registration"""
        DUMMYNODEID = "90f7c2c0-cfa9-11e7-9b9d-2fe338e1e7ce"
        a = Aggregator(mdns_updater=mock.MagicMock())
        a._node_data["registered"] = True
        a._node_data["node"] = {"type": "node", "data": {"id": DUMMYNODEID}}
        a.aggregator = "http://example.com"

        with mock.patch.object(a, '_send', side_effect=InvalidRequest(status_code=404)) as send:
            with mock.patch.object(a, '_register_node', return_value=True) as register:
                return_val = a._heartbeat()

                self.assertTrue(return_val)
                send.assert_called_with("POST", "http://example.com", a.aggregator_apiversion,
                                        "health/nodes/{}".format(DUMMYNODEID))
                a._mdns_updater.inc_P2P_enable_count.assert_not_called()
                register.assert_called_once_with(a._node_data["node"])
                self.assertFalse(a._node_data["registered"])

    def test_heartbeat_404_failure(self):
        """Test heartbeat operation when heartbeat request returns HTTP 404 and aggregator in failed state
        after HTTP 404, Node should mark itself as not registered and attempt re-registration"""
        DUMMYNODEID = "90f7c2c0-cfa9-11e7-9b9d-2fe338e1e7ce"
        a = Aggregator(mdns_updater=mock.MagicMock())
        a._node_data["registered"] = True
        a._node_data["node"] = {"type": "node", "data": {"id": DUMMYNODEID}}
        a.aggregator = "http://example.com"

        with mock.patch.object(a, '_send', side_effect=InvalidRequest(status_code=404)) as send:
            with mock.patch.object(a, '_register_node', return_value=False) as register:
                return_val = a._heartbeat()

                self.assertFalse(return_val)
                send.assert_called_with("POST", "http://example.com", a.aggregator_apiversion,
                                        "health/nodes/{}".format(DUMMYNODEID))
                a._mdns_updater.inc_P2P_enable_count.assert_not_called()
                register.assert_called_once_with(a._node_data["node"])
                self.assertFalse(a._node_data["registered"])

    def test_heartbeat_4xx_failure(self):
        """Test heartbeat operation when heartbeat request returns HTTP 4xx and aggregator in failed state
        Indicates aggregator has encountered an error, heartbeat should fail and next aggregator should be used"""
        DUMMYNODEID = "90f7c2c0-cfa9-11e7-9b9d-2fe338e1e7ce"
        a = Aggregator(mdns_updater=mock.MagicMock())
        a._node_data["registered"] = True
        a._node_data["node"] = {"type": "node", "data": {"id": DUMMYNODEID}}
        a.aggregator = "http://example.com"

        with mock.patch.object(a, '_send', side_effect=InvalidRequest(status_code=401)) as send:
            return_val = a._heartbeat()

            self.assertFalse(return_val)
            send.assert_called_with("POST", "http://example.com", a.aggregator_apiversion,
                                    "health/nodes/{}".format(DUMMYNODEID))
            a._mdns_updater.inc_P2P_enable_count.assert_not_called()
            self.assertTrue(a._node_data["registered"])

    def test_heartbeat_5xx_failure(self):
        """Test heartbeat operation when heartbeat request returns HTTP 5xx and aggregator in failed state
        Indicates aggregator has encountered an error, heartbeat should fail and next aggregator should be used"""
        DUMMYNODEID = "90f7c2c0-cfa9-11e7-9b9d-2fe338e1e7ce"
        a = Aggregator(mdns_updater=mock.MagicMock())
        a._node_data["registered"] = True
        a._node_data["node"] = {"type": "node", "data": {"id": DUMMYNODEID}}
        a.aggregator = "http://example.com"

        with mock.patch.object(a, '_send', side_effect=ServerSideError) as send:
            return_val = a._heartbeat()

            self.assertFalse(return_val)
            send.assert_called_with("POST", "http://example.com", a.aggregator_apiversion,
                                    "health/nodes/{}".format(DUMMYNODEID))
            a._mdns_updater.inc_P2P_enable_count.assert_not_called()
            self.assertTrue(a._node_data["registered"])

    def test_heartbeat_exception_failure(self):
        """Test heartbeat operation when heartbeat request returns an Exception
        Indicates aggregator has encountered an error, heartbeat should fail and next aggregator should be used"""
        DUMMYNODEID = "90f7c2c0-cfa9-11e7-9b9d-2fe338e1e7ce"
        a = Aggregator(mdns_updater=mock.MagicMock())
        a._node_data["registered"] = True
        a._node_data["node"] = {"type": "node", "data": {"id": DUMMYNODEID}}
        a.aggregator = "http://example.com"

        with mock.patch.object(a, '_send', side_effect=Exception) as send:
            return_val = a._heartbeat()

            self.assertFalse(return_val)
            send.assert_called_with("POST", "http://example.com", a.aggregator_apiversion,
                                    "health/nodes/{}".format(DUMMYNODEID))
            a._mdns_updater.inc_P2P_enable_count.assert_not_called()
            self.assertFalse(a._node_data["registered"])

    def test_heartbeat_if_no_aggregator_set(self):
        """Test heartbeat operation when no aggregator set
        No heartbeat should be performed and should return False"""
        DUMMYNODEID = "90f7c2c0-cfa9-11e7-9b9d-2fe338e1e7ce"
        a = Aggregator(mdns_updater=mock.MagicMock())
        a._node_data["registered"] = True
        a._node_data["node"] = {"type": "node", "data": {"id": DUMMYNODEID}}
        a.aggregator = None

        with mock.patch.object(a, '_send', side_effect=Exception) as send:
            return_val = a._heartbeat()

            self.assertFalse(return_val)
            send.assert_not_called()
            a._mdns_updater.inc_P2P_enable_count.assert_not_called()
            self.assertTrue(a._node_data["registered"])

    # # ================================================================================================================
    # # Test discovery operation
    # # ================================================================================================================

    def test_discovery_under_normal_operation(self):
        """Test the discovery correctly registers node with aggregator and heartbeats"""
        BACKOFF_PERIOD = 4
        AGGREGATOR_1 = "http://example1.com"

        a = Aggregator(mdns_updater=mock.MagicMock())
        a._node_data["registered"] = False
        a.aggregator = None
        a._aggregator_list_stale = True
        a._backoff_period = BACKOFF_PERIOD

        a.mdnsbridge.getHrefWithException.return_value = AGGREGATOR_1

        with mock.patch('gevent.sleep') as sleep:
            with mock.patch.object(a, '_heartbeat', return_value=True) as heartbeat:
                a._discovery_operation()

                sleep.assert_called_once_with(BACKOFF_PERIOD)
                a.mdnsbridge.updateServices.assert_called_once_with(LEGACY_REG_MDNSTYPE)
                heartbeat.assert_called_once_with()
                self.assertFalse(a._backoff_active)
                self.assertFalse(a._aggregator_list_stale)
                self.assertEqual(a.aggregator, AGGREGATOR_1)
                a._mdns_updater.inc_P2P_enable_count.assert_not_called()

    def test_discovery_when_aggregator_failure_flag_set(self):
        """Test the discovery correctly registers node with aggregator and heartbeats,
        but does not when for backoff period"""
        AGGREGATOR_1 = "http://example1.com"

        a = Aggregator(mdns_updater=mock.MagicMock())
        a._node_data["registered"] = False
        a.aggregator = None
        a._aggregator_list_stale = True
        a._aggregator_failure = True

        a.mdnsbridge.getHrefWithException.return_value = AGGREGATOR_1

        with mock.patch('gevent.sleep') as sleep:
            with mock.patch.object(a, '_heartbeat', return_value=True) as heartbeat:
                a._discovery_operation()

                sleep.assert_not_called()
                a.mdnsbridge.updateServices.assert_called_once_with(LEGACY_REG_MDNSTYPE)
                heartbeat.assert_called_once_with()
                self.assertFalse(a._backoff_active)
                self.assertFalse(a._aggregator_list_stale)
                self.assertEqual(a.aggregator, AGGREGATOR_1)
                a._mdns_updater.inc_P2P_enable_count.assert_not_called()

    def test_discovery_when_no_aggregator_returned(self):
        """Test the discovery returns when no aggregators found and does not attempt registration"""

        a = Aggregator(mdns_updater=mock.MagicMock())
        a._node_data["registered"] = False
        a.aggregator = None
        a._aggregator_list_stale = True
        a._aggregator_failure = True

        a.mdnsbridge.getHrefWithException.side_effect = NoService

        with mock.patch('gevent.sleep') as sleep:
            with mock.patch.object(a, '_heartbeat', return_value=True) as heartbeat:
                a._discovery_operation()

                sleep.assert_not_called()
                a.mdnsbridge.updateServices.assert_called_once_with(LEGACY_REG_MDNSTYPE)
                heartbeat.assert_not_called()
                self.assertFalse(a._backoff_active)
                self.assertTrue(a._aggregator_list_stale)
                self.assertEqual(a.aggregator, None)
                a._mdns_updater.inc_P2P_enable_count.assert_called_once_with()
                self.assertEqual(a._backoff_period, BACKOFF_INITIAL_TIMOUT_SECONDS)

    def test_discovery_when_end_of_list_returned(self):
        """Test the discovery returns when end of aggregator list reached and does not attempt registration"""

        a = Aggregator(mdns_updater=mock.MagicMock())
        a._node_data["registered"] = False
        a.aggregator = None
        a._aggregator_list_stale = True
        a._aggregator_failure = True

        a.mdnsbridge.getHrefWithException.side_effect = EndOfServiceList

        with mock.patch('gevent.sleep') as sleep:
            with mock.patch.object(a, '_heartbeat', return_value=True) as heartbeat:
                a._discovery_operation()

                sleep.assert_not_called()
                a.mdnsbridge.updateServices.assert_called_once_with(LEGACY_REG_MDNSTYPE)
                heartbeat.assert_not_called()
                self.assertFalse(a._backoff_active)
                self.assertTrue(a._aggregator_list_stale)
                self.assertEqual(a.aggregator, None)
                a._mdns_updater.inc_P2P_enable_count.assert_not_called()
                self.assertEqual(a._backoff_period, BACKOFF_INITIAL_TIMOUT_SECONDS)

    def test_discovery_when_heartbeat_fails(self):
        """Test that discovery uses tries multiple aggregators before returning"""
        AGGREGATOR_1 = "http://example1.com"
        AGGREGATOR_2 = "http://example2.com"

        a = Aggregator(mdns_updater=mock.MagicMock())
        a._node_data["registered"] = False
        a.aggregator = None
        a._aggregator_list_stale = True

        a.mdnsbridge.getHrefWithException.side_effect = [AGGREGATOR_1, AGGREGATOR_2]

        with mock.patch('gevent.sleep'):
            with mock.patch.object(a, '_heartbeat', side_effect=[False, True]) as heartbeat:
                a._discovery_operation()

                a.mdnsbridge.updateServices.assert_called_once_with(LEGACY_REG_MDNSTYPE)
                heartbeat.assert_called_with()
                self.assertFalse(a._backoff_active)
                self.assertFalse(a._aggregator_list_stale)
                self.assertEqual(a.aggregator, AGGREGATOR_2)
                a._mdns_updater.inc_P2P_enable_count.assert_not_called()

    # # ================================================================================================================
    # # Test registered operation
    # # ================================================================================================================

    def test_registered_operation_under_normal_operation(self):
        """Test the registered operation correctly heartbeats"""
        AGGREGATOR_1 = "http://example1.com"

        a = Aggregator(mdns_updater=mock.MagicMock())
        a._node_data["registered"] = False
        a.aggregator = AGGREGATOR_1

        with mock.patch.object(a, '_heartbeat', return_value=True) as heartbeat:
            a._registered_operation()

            heartbeat.assert_called_once_with()
            self.assertFalse(a._backoff_active)
            self.assertFalse(a._aggregator_failure)
            self.assertEqual(a.aggregator, AGGREGATOR_1)
            a._mdns_updater.inc_P2P_enable_count.assert_not_called()

    def test_registered_operation_when_heartbeat_fails(self):
        """Test the registered operation when the heartbeat fails"""
        AGGREGATOR_1 = "http://example1.com"

        a = Aggregator(mdns_updater=mock.MagicMock())
        a._node_data["registered"] = False
        a.aggregator = AGGREGATOR_1

        with mock.patch.object(a, '_heartbeat', return_value=False) as heartbeat:
            a._registered_operation()

            heartbeat.assert_called_once_with()
            self.assertFalse(a._backoff_active)
            self.assertTrue(a._aggregator_failure)
            self.assertEqual(a.aggregator, None)
            a._mdns_updater.inc_P2P_enable_count.assert_not_called()

    # # ================================================================================================================
    # # Test _register_node
    # # ================================================================================================================

    def test_register_node_under_normal_operation(self):
        """Test that the Node correctly registers with aggregator"""
        DUMMYNODEID = "90f7c2c0-cfa9-11e7-9b9d-2fe338e1e7ce"
        AGGREGATOR_1 = "http://example1.com"

        a = Aggregator(mdns_updater=mock.MagicMock())
        a._node_data["registered"] = False
        a.aggregator = AGGREGATOR_1
        a._node_data["node"] = {"type": "node", "data": {"id": DUMMYNODEID}}

        def request(*args, **kwargs):
            return mock.MagicMock(status_code=201)

        with mock.patch.object(a, '_send', side_effect=request) as send:
            with mock.patch.object(a, '_register_node_resources') as reg_resources:
                return_val = a._register_node(a._node_data["node"])

                self.assertTrue(return_val)
                send.assert_called_once_with("POST", AGGREGATOR_1, a.aggregator_apiversion,
                                             "resource", a._node_data["node"])
                reg_resources.assert_called_once_with()
                self.assertFalse(a._backoff_active)
                self.assertTrue(a._aggregator_list_stale)
                self.assertTrue(a._node_data['registered'])
                a._mdns_updater.inc_P2P_enable_count.assert_not_called()

    def test_register_node_when_no_node_data(self):
        """Test that is there is no data the registration fails"""
        a = Aggregator(mdns_updater=mock.MagicMock())
        a._node_data["registered"] = False

        with mock.patch.object(a, '_send') as send:
            with mock.patch.object(a, '_register_node_resources') as reg_resources:
                return_val = a._register_node(None)

                self.assertFalse(return_val)
                send.assert_not_called()
                reg_resources.assert_not_called()
                self.assertFalse(a._backoff_active)
                self.assertTrue(a._aggregator_list_stale)
                self.assertFalse(a._node_data['registered'])
                a._mdns_updater.inc_P2P_enable_count.assert_not_called()

    def test_register_node_response_200(self):
        """Test that the Node correctly un-registers with aggregator when HTTP 200 response received"""
        DUMMYNODEID = "90f7c2c0-cfa9-11e7-9b9d-2fe338e1e7ce"
        AGGREGATOR_1 = "http://example1.com"

        a = Aggregator(mdns_updater=mock.MagicMock())
        a._node_data["registered"] = False
        a.aggregator = AGGREGATOR_1
        a._node_data["node"] = {"type": "node", "data": {"id": DUMMYNODEID}}

        request_list = [
            mock.MagicMock(status_code=200, headers={'Location': 'path/xxx'}),
            mock.MagicMock(status_code=201),
        ]

        with mock.patch.object(a, '_send', side_effect=request_list) as send:
            with mock.patch.object(a, '_unregister_node', return_value=True) as un_reg:
                with mock.patch.object(a, '_register_node_resources') as reg_resources:
                    return_val = a._register_node(a._node_data["node"])

                    self.assertTrue(return_val)
                    send.assert_called_with("POST", AGGREGATOR_1, a.aggregator_apiversion,
                                            "resource", a._node_data["node"])
                    un_reg.assert_called_once_with('path/xxx')
                    reg_resources.assert_called_once_with()
                    self.assertFalse(a._backoff_active)
                    self.assertTrue(a._aggregator_list_stale)
                    self.assertTrue(a._node_data['registered'])
                    a._mdns_updater.inc_P2P_enable_count.assert_not_called()

    def test_register_node_response_409(self):
        """Test that the Node correctly un-registers with aggregator when HTTP 409 response received"""
        DUMMYNODEID = "90f7c2c0-cfa9-11e7-9b9d-2fe338e1e7ce"
        AGGREGATOR_1 = "http://example1.com"

        a = Aggregator(mdns_updater=mock.MagicMock())
        a._node_data["registered"] = False
        a.aggregator = AGGREGATOR_1
        a._node_data["node"] = {"type": "node", "data": {"id": DUMMYNODEID}}

        request_list = [
            mock.MagicMock(status_code=409, headers={'Location': 'path/xxx'}),
            mock.MagicMock(status_code=201),
        ]

        with mock.patch.object(a, '_send', side_effect=request_list) as send:
            with mock.patch.object(a, '_unregister_node', return_value=True) as un_reg:
                with mock.patch.object(a, '_register_node_resources') as reg_resources:
                    return_val = a._register_node(a._node_data["node"])

                    self.assertTrue(return_val)
                    send.assert_called_with("POST", AGGREGATOR_1, a.aggregator_apiversion,
                                            "resource", a._node_data["node"])
                    un_reg.assert_called_once_with('path/xxx')
                    reg_resources.assert_called_once_with()
                    self.assertFalse(a._backoff_active)
                    self.assertTrue(a._aggregator_list_stale)
                    self.assertTrue(a._node_data['registered'])
                    a._mdns_updater.inc_P2P_enable_count.assert_not_called()

    def test_register_node_response_Exception(self):
        """Test that the Node correctly handles Exception during registration"""
        DUMMYNODEID = "90f7c2c0-cfa9-11e7-9b9d-2fe338e1e7ce"
        AGGREGATOR_1 = "http://example1.com"

        a = Aggregator(mdns_updater=mock.MagicMock())
        a._node_data["registered"] = False
        a.aggregator = AGGREGATOR_1
        a._node_data["node"] = {"type": "node", "data": {"id": DUMMYNODEID}}

        with mock.patch.object(a, '_send', side_effect=Exception) as send:
            return_val = a._register_node(a._node_data["node"])

            self.assertFalse(return_val)
            send.assert_called_with("POST", AGGREGATOR_1, a.aggregator_apiversion,
                                    "resource", a._node_data["node"])
            self.assertFalse(a._backoff_active)
            self.assertFalse(a._node_data['registered'])
            a._mdns_updater.inc_P2P_enable_count.assert_not_called()

    def test_register_node_response_if_un_registration_fails(self):
        """Test that the Node correctly handles a failure during un-registration"""
        DUMMYNODEID = "90f7c2c0-cfa9-11e7-9b9d-2fe338e1e7ce"
        AGGREGATOR_1 = "http://example1.com"

        a = Aggregator(mdns_updater=mock.MagicMock())
        a._node_data["registered"] = False
        a.aggregator = AGGREGATOR_1
        a._node_data["node"] = {"type": "node", "data": {"id": DUMMYNODEID}}

        request_response = [
            mock.MagicMock(status_code=409, headers={'Location': 'path/xxx'})
        ]

        with mock.patch.object(a, '_send', side_effect=request_response) as send:
            with mock.patch.object(a, '_unregister_node', return_value=False) as un_reg:

                return_val = a._register_node(a._node_data["node"])

                self.assertFalse(return_val)
                send.assert_called_with("POST", AGGREGATOR_1, a.aggregator_apiversion,
                                        "resource", a._node_data["node"])
                un_reg.assert_called_once_with('path/xxx')
                self.assertFalse(a._backoff_active)
                self.assertFalse(a._node_data['registered'])
                a._mdns_updater.inc_P2P_enable_count.assert_not_called()

    def test_register_node_drains_queue(self):
        """Test that the register Node drains the queue before registration"""
        DUMMYNODEID = "90f7c2c0-cfa9-11e7-9b9d-2fe338e1e7ce"
        AGGREGATOR_1 = "http://example1.com"

        a = Aggregator(mdns_updater=mock.MagicMock())
        a._node_data["registered"] = False
        a.aggregator = AGGREGATOR_1
        a._node_data["node"] = {"type": "node", "data": {"id": DUMMYNODEID}}

        request_response = [
            mock.MagicMock(status_code=201)
        ]

        queue = [
            {"method": "POST", "namespace": "resource", "res_type": "dummy", "key": DUMMYNODEID},
            {"method": "DELETE", "namespace": "resource", "res_type": "dummy", "key": DUMMYNODEID}
        ]

        a._reg_queue.empty.side_effect = lambda: (len(queue) == 0)

        def _get(block=True):
            if len(queue) == 0:
                raise gevent.queue.Queue.Empty
            return queue.pop(0)
        a._reg_queue.get.side_effect = _get

        with mock.patch.object(a, '_send', side_effect=request_response) as send:
            with mock.patch.object(a, '_register_node_resources') as reg_resources:
                return_val = a._register_node(a._node_data["node"])

                self.assertTrue(return_val)
                send.assert_called_with("POST", AGGREGATOR_1, a.aggregator_apiversion,
                                        "resource", a._node_data["node"])
                reg_resources.assert_called_once_with()
                self.assertFalse(a._backoff_active)
                self.assertTrue(a._node_data['registered'])
                a._mdns_updater.inc_P2P_enable_count.assert_not_called()

    # # ================================================================================================================
    # # Test _unregister_node
    # # ================================================================================================================

    def test_unregister_node_under_normal_operation(self):
        """Test that the Node correctly unregisters with aggregator"""
        DUMMYNODEID = "90f7c2c0-cfa9-11e7-9b9d-2fe338e1e7ce"
        AGGREGATOR_1 = "http://example1.com:234"

        a = Aggregator(mdns_updater=mock.MagicMock())
        a._node_data["registered"] = True
        a.aggregator = AGGREGATOR_1
        a._node_data["node"] = {"type": "node", "data": {"id": DUMMYNODEID}}

        def request(*args, **kwargs):
            return mock.MagicMock(status_code=204)

        # No URL Path
        with mock.patch.object(a, '_send', side_effect=request) as send:
            return_val = a._unregister_node(None)

            self.assertTrue(return_val)
            send.assert_called_once_with("DELETE", AGGREGATOR_1, a.aggregator_apiversion,
                                         "resource/nodes/{}".format(DUMMYNODEID))
            self.assertFalse(a._node_data['registered'])

        # Relative URL
        a._node_data["registered"] = True
        with mock.patch.object(a, '_send_request', side_effect=request) as send:
            return_val = a._unregister_node('/path/xxx')

            self.assertTrue(return_val)
            send.assert_called_once_with("DELETE", AGGREGATOR_1, '/path/xxx')
            self.assertFalse(a._node_data['registered'])

        # Absolute URL
        a._node_data["registered"] = True
        with mock.patch.object(a, '_send_request', side_effect=request) as send:
            return_val = a._unregister_node(AGGREGATOR_1 + '/path/xxx')

            self.assertTrue(return_val)
            send.assert_called_once_with("DELETE", AGGREGATOR_1, '/path/xxx')
            self.assertFalse(a._node_data['registered'])

    def test_unregister_node_unexpected_response(self):
        """Test that the Node returns False when DELETE returns an unexpected response"""
        DUMMYNODEID = "90f7c2c0-cfa9-11e7-9b9d-2fe338e1e7ce"
        AGGREGATOR_1 = "http://example1.com"

        a = Aggregator(mdns_updater=mock.MagicMock())
        a._node_data["registered"] = True
        a.aggregator = AGGREGATOR_1
        a._node_data["node"] = {"type": "node", "data": {"id": DUMMYNODEID}}

        def request(*args, **kwargs):
            return mock.MagicMock(status_code=200)

        with mock.patch.object(a, '_send', side_effect=request) as send:
            return_val = a._unregister_node(None)

            self.assertFalse(return_val)
            send.assert_called_once_with("DELETE", AGGREGATOR_1, a.aggregator_apiversion,
                                         "resource/nodes/{}".format(DUMMYNODEID))
            self.assertFalse(a._node_data['registered'])

    def test_unregister_node_Exception(self):
        """Test that the Node returns False when Exception raise during DELETE"""
        DUMMYNODEID = "90f7c2c0-cfa9-11e7-9b9d-2fe338e1e7ce"
        AGGREGATOR_1 = "http://example1.com"

        a = Aggregator(mdns_updater=mock.MagicMock())
        a._node_data["registered"] = True
        a.aggregator = AGGREGATOR_1
        a._node_data["node"] = {"type": "node", "data": {"id": DUMMYNODEID}}

        with mock.patch.object(a, '_send', side_effect=Exception) as send:
            return_val = a._unregister_node(None)

            self.assertFalse(return_val)
            send.assert_called_once_with("DELETE", AGGREGATOR_1, a.aggregator_apiversion,
                                         "resource/nodes/{}".format(DUMMYNODEID))
            self.assertFalse(a._node_data['registered'])

    # # ================================================================================================================
    # # Test _register_node_resources
    # # ================================================================================================================

    def test_register_node_resources_under_normal_operation(self):
        """Test that the Node correctly queues resources to be registered with aggregator"""
        DUMMYNODEID = "90f7c2c0-cfa9-11e7-9b9d-2fe338e1e7ce"
        DUMMYKEY = "dummykey"
        DUMMYPARAMKEY = "dummyparamkey"
        DUMMYPARAMVAL = "dummyparamval"
        DUMMYFLOW = "dummyflow"
        DUMMYDEVICE = "dummydevice"
        AGGREGATOR_1 = "http://example1.com"

        a = Aggregator(mdns_updater=mock.MagicMock())
        a._node_data["registered"] = True
        a.aggregator = AGGREGATOR_1
        a._node_data["node"] = {"type": "node", "data": {"id": DUMMYNODEID}}
        if "entities" not in a._node_data:
            a._node_data["entities"] = {}
        if "resource" not in a._node_data["entities"]:
            a._node_data["entities"]["resource"] = {}
        if "dummy" not in a._node_data["entities"]["resource"]:
            a._node_data["entities"]["resource"]["dummy"] = {}
        if "device" not in a._node_data["entities"]["resource"]:
            a._node_data["entities"]["resource"]["device"] = {}
        if "flow" not in a._node_data["entities"]["resource"]:
            a._node_data["entities"]["resource"]["flow"] = {}
        a._node_data["entities"]["resource"]["dummy"][DUMMYKEY] = {DUMMYPARAMKEY: DUMMYPARAMVAL}
        a._node_data["entities"]["resource"]["device"][DUMMYDEVICE] = {DUMMYPARAMKEY: DUMMYPARAMVAL}
        a._node_data["entities"]["resource"]["flow"][DUMMYFLOW] = {DUMMYPARAMKEY: DUMMYPARAMVAL}

        expected_put_calls = []
        # The re-registration of the other resources should be queued for the next run loop, and arranged in order
        expected_put_calls = (
            sum([
                [mock.call({"method": "POST", "namespace": "resource", "res_type": res_type, "key": key})
                    for key in a._node_data["entities"]["resource"][res_type]
                 ] for res_type in a.registration_order if res_type in a._node_data["entities"]["resource"]
            ], []) + sum([
                [mock.call({"method": "POST", "namespace": "resource", "res_type": res_type, "key": key})
                    for key in a._node_data["entities"]["resource"][res_type]
                 ] for res_type in a._node_data["entities"]["resource"] if res_type not in a.registration_order
            ], [])
        )

        a._register_node_resources()
        self.assertListEqual(a._reg_queue.put.mock_calls, expected_put_calls)

    # # ================================================================================================================
    # # Test queue handelling
    # # ================================================================================================================

    def test_process_queue_does_nothing_when_not_registered(self):
        """The queue processing thread should not send any messages when the node is not registered."""
        a = Aggregator(mdns_updater=mock.MagicMock())
        a._node_data["registered"] = False
        a._reg_queue.empty.return_value = False

        def killloop(*args, **kwargs):
            a._running = False

        with mock.patch('gevent.sleep', side_effect=killloop) as sleep:
            with mock.patch.object(a, '_send') as send:
                a._process_queue()

                send.assert_not_called()
                a._mdns_updater.P2P_disable.assert_not_called()
                sleep.assert_called_with(mock.ANY)

    def test_process_queue_does_nothing_when_queue_empty(self):
        """The queue processing thread should not send any messages when the queue is empty."""
        DUMMYNODEID = "90f7c2c0-cfa9-11e7-9b9d-2fe338e1e7ce"
        a = Aggregator(mdns_updater=mock.MagicMock())
        a._node_data["registered"] = True
        a._node_data["node"] = {"type": "node", "data": {"id": DUMMYNODEID}}
        a._reg_queue.empty.return_value = True

        def killloop(*args, **kwargs):
            a._running = False

        with mock.patch('gevent.sleep', side_effect=killloop) as sleep:
            with mock.patch.object(a, '_send') as send:
                a._process_queue()

                send.assert_not_called()
                a._mdns_updater.P2P_disable.assert_not_called()
                sleep.assert_called_with(mock.ANY)

    def test_process_queue_processes_queue_when_running(self):
        """The queue processing thread should check the queue and send a registration/deregistration request
        to the remote aggregator when required."""
        DUMMYNODEID = "90f7c2c0-cfa9-11e7-9b9d-2fe338e1e7ce"
        DUMMYKEY = "dummykey"
        DUMMYPARAMKEY = "dummyparamkey"
        DUMMYPARAMVAL = "dummyparamval"

        a = Aggregator(mdns_updater=mock.MagicMock())
        a.aggregator = "www.example.com"
        a._node_data["registered"] = True
        a._node_data["node"] = {"type": "node", "data": {"id": DUMMYNODEID}}
        if "entities" not in a._node_data:
            a._node_data["entities"] = {}
        if "resource" not in a._node_data["entities"]:
            a._node_data["entities"]["resource"] = {}
        if "dummy" not in a._node_data["entities"]["resource"]:
            a._node_data["entities"]["resource"]["dummy"] = {}
        a._node_data["entities"]["resource"]["dummy"][DUMMYKEY] = {DUMMYPARAMKEY: DUMMYPARAMVAL}

        queue = [
            {"method": "POST", "namespace": "resource", "res_type": "node", "key": DUMMYNODEID},
            {"method": "POST", "namespace": "resource", "res_type": "dummy", "key": DUMMYKEY},
            {"method": "DELETE", "namespace": "resource", "res_type": "dummy", "key": DUMMYKEY}
        ]

        a._reg_queue.empty.side_effect = lambda: (len(queue) == 0)
        a._reg_queue.get.side_effect = lambda: queue.pop(0)

        expected_calls = [
            mock.call('POST', "www.example.com", "v1.2", 'resource',
                      a._node_data["node"]),
            mock.call('POST', "www.example.com", "v1.2", 'resource',
                      a._node_data["entities"]["resource"]["dummy"][DUMMYKEY]),
            mock.call('DELETE', "www.example.com", "v1.2", 'resource/dummys/' + DUMMYKEY)
        ]

        def killloop(*args, **kwargs):
            a._running = False

        with mock.patch('gevent.sleep', side_effect=killloop):
            with mock.patch.object(a, '_send') as send:
                a._process_queue()
                send.assert_has_calls(expected_calls)

    def test_process_queue_stops_when_not_running(self):
        """The process queue method should stop as soon as running set to false"""
        DUMMYNODEID = "90f7c2c0-cfa9-11e7-9b9d-2fe338e1e7ce"
        DUMMYKEY = "dummykey"
        DUMMYPARAMKEY = "dummyparamkey"
        DUMMYPARAMVAL = "dummyparamval"

        a = Aggregator(mdns_updater=mock.MagicMock())
        a._running = False
        a.aggregator = "www.example.com"
        a._node_data["registered"] = True
        a._node_data["node"] = {"type": "node", "data": {"id": DUMMYNODEID}}
        if "entities" not in a._node_data:
            a._node_data["entities"] = {}
        if "resource" not in a._node_data["entities"]:
            a._node_data["entities"]["resource"] = {}
        if "dummy" not in a._node_data["entities"]["resource"]:
            a._node_data["entities"]["resource"]["dummy"] = {}
        a._node_data["entities"]["resource"]["dummy"][DUMMYKEY] = {DUMMYPARAMKEY: DUMMYPARAMVAL}

        queue = [
            {"method": "POST", "namespace": "resource", "res_type": "node", "key": DUMMYNODEID},
        ]

        a._reg_queue.empty.side_effect = lambda: (len(queue) == 0)
        a._reg_queue.get.side_effect = lambda: queue.pop(0)

        with mock.patch('gevent.sleep', side_effect=Exception) as sleep:
            with mock.patch.object(a, '_send') as send:
                try:
                    a._process_queue()
                except Exception:
                    self.fail(msg="process_queue kept running")

                send.assert_not_called
                sleep.assert_not_called()

    def test_process_queue_processes_queue_when_running_and_aborts_on_exception_in_general_register(self):
        """If a non-node register performed by the queue processing thread throws an exception then the loop
        should abort."""
        DUMMYNODEID = "90f7c2c0-cfa9-11e7-9b9d-2fe338e1e7ce"
        DUMMYKEY = "dummykey"
        DUMMYPARAMKEY = "dummyparamkey"
        DUMMYPARAMVAL = "dummyparamval"

        a = Aggregator(mdns_updater=mock.MagicMock())
        a.aggregator = "www.example.com"
        a._node_data["registered"] = True
        a._node_data["node"] = {"type": "node", "data": {"id": DUMMYNODEID}}
        if "entities" not in a._node_data:
            a._node_data["entities"] = {}
        if "resource" not in a._node_data["entities"]:
            a._node_data["entities"]["resource"] = {}
        if "dummy" not in a._node_data["entities"]["resource"]:
            a._node_data["entities"]["resource"]["dummy"] = {}
        a._node_data["entities"]["resource"]["dummy"][DUMMYKEY] = {DUMMYPARAMKEY: DUMMYPARAMVAL}

        queue = [
            {"method": "POST", "namespace": "resource", "res_type": "dummy", "key": DUMMYKEY},
        ]

        a._reg_queue.empty.side_effect = lambda: (len(queue) == 0)
        a._reg_queue.get.side_effect = lambda: queue.pop(0)

        expected_calls = [
            mock.call('POST', "www.example.com", "v1.2", 'resource',
                      a._node_data["entities"]["resource"]["dummy"][DUMMYKEY]),
        ]

        def killloop(*args, **kwargs):
            a._running = False

        with mock.patch('gevent.sleep', side_effect=killloop):
            with mock.patch.object(a, '_send', side_effect=InvalidRequest) as send:
                a._process_queue()

                send.assert_has_calls(expected_calls)
                a._mdns_updater.P2P_disable.assert_not_called()
                self.assertNotIn(DUMMYKEY, a._node_data["entities"]["resource"]["dummy"])

    def test_process_queue_processes_queue_when_running_and_aborts_on_exception_in_general_unregister(self):
        """If an unregister performed by the queue processing thread throws an exception then the loop should abort."""
        DUMMYNODEID = "90f7c2c0-cfa9-11e7-9b9d-2fe338e1e7ce"
        DUMMYKEY = "dummykey"
        DUMMYPARAMKEY = "dummyparamkey"
        DUMMYPARAMVAL = "dummyparamval"

        a = Aggregator(mdns_updater=mock.MagicMock())
        a.aggregator = "www.example.com"
        a._node_data["registered"] = True
        a._node_data["node"] = {"type": "node", "data": {"id": DUMMYNODEID}}
        if "entities" not in a._node_data:
            a._node_data["entities"] = {}
        if "resource" not in a._node_data["entities"]:
            a._node_data["entities"]["resource"] = {}
        if "dummy" not in a._node_data["entities"]["resource"]:
            a._node_data["entities"]["resource"]["dummy"] = {}
        a._node_data["entities"]["resource"]["dummy"][DUMMYKEY] = {DUMMYPARAMKEY: DUMMYPARAMVAL}

        queue = [
            {"method": "DELETE", "namespace": "resource", "res_type": "dummy", "key": DUMMYKEY}
        ]

        a._reg_queue.empty.side_effect = lambda: (len(queue) == 0)
        a._reg_queue.get.side_effect = lambda: queue.pop(0)

        expected_calls = [
            mock.call('DELETE', "www.example.com", "v1.2", 'resource/dummys/' + DUMMYKEY)
        ]

        def killloop(*args, **kwargs):
            a._running = False

        with mock.patch('gevent.sleep', side_effect=killloop):
            with mock.patch.object(a, '_send', side_effect=InvalidRequest) as send:
                a._process_queue()

                send.assert_has_calls(expected_calls)
                a._mdns_updater.P2P_disable.assert_not_called()

    def test_process_queue_processes_queue_when_running_and_aborts_on_server_side_exception(self):
        """If a non-node register performed by the queue processing thread throws a server side exception,
        new aggregator should be chosen and failed request should be queued added to the front of the queue."""
        DUMMYNODEID = "90f7c2c0-cfa9-11e7-9b9d-2fe338e1e7ce"
        DUMMYKEY = "dummykey"
        DUMMYPARAMKEY = "dummyparamkey"
        DUMMYPARAMVAL = "dummyparamval"

        a = Aggregator(mdns_updater=mock.MagicMock())
        a.aggregator = "www.example.com"
        a._node_data["registered"] = True
        a._node_data["node"] = {"type": "node", "data": {"id": DUMMYNODEID}}
        if "entities" not in a._node_data:
            a._node_data["entities"] = {}
        if "resource" not in a._node_data["entities"]:
            a._node_data["entities"]["resource"] = {}
        if "dummy" not in a._node_data["entities"]["resource"]:
            a._node_data["entities"]["resource"]["dummy"] = {}
        a._node_data["entities"]["resource"]["dummy"][DUMMYKEY] = {DUMMYPARAMKEY: DUMMYPARAMVAL}

        expected_queue = [
            {"method": "POST", "namespace": "resource", "res_type": "dummy", "key": DUMMYKEY},
            {"method": "POST", "namespace": "resource", "res_type": "left", "key": DUMMYKEY},
        ]
        queue = deepcopy(expected_queue)
        updated_queue = []

        a._reg_queue.empty.side_effect = lambda: (len(queue) == 0)
        a._reg_queue.get.side_effect = lambda: queue.pop(0)
        a._reg_queue.put.side_effect = lambda data: updated_queue.append(data)

        expected_calls = [
            mock.call('POST', "www.example.com", "v1.2", 'resource',
                      a._node_data["entities"]["resource"]["dummy"][DUMMYKEY]),
        ]

        def killloop(*args, **kwargs):
            a._running = False

        with mock.patch('gevent.sleep', side_effect=killloop):
            with mock.patch.object(a, '_send', side_effect=ServerSideError) as send:
                a._process_queue()

                send.assert_has_calls(expected_calls)
                a._mdns_updater.P2P_disable.assert_not_called()
                self.assertIsNone(a.aggregator)
                self.assertListEqual(updated_queue, expected_queue)

    def test_process_queue_processes_queue_when_running_and_ignores_unknown_methods(self):
        """Unknown verbs in the queue should be ignored."""
        DUMMYNODEID = "90f7c2c0-cfa9-11e7-9b9d-2fe338e1e7ce"
        DUMMYKEY = "dummykey"
        DUMMYPARAMKEY = "dummyparamkey"
        DUMMYPARAMVAL = "dummyparamval"

        a = Aggregator(mdns_updater=mock.MagicMock())
        a.aggregator = "www.example.com"
        a._node_data["registered"] = True
        a._node_data["node"] = {"type": "node", "data": {"id": DUMMYNODEID}}
        if "entities" not in a._node_data:
            a._node_data["entities"] = {}
        if "resource" not in a._node_data["entities"]:
            a._node_data["entities"]["resource"] = {}
        if "dummy" not in a._node_data["entities"]["resource"]:
            a._node_data["entities"]["resource"]["dummy"] = {}
        a._node_data["entities"]["resource"]["dummy"][DUMMYKEY] = {DUMMYPARAMKEY: DUMMYPARAMVAL}

        queue = [
            {"method": "DANCE", "namespace": "resource", "res_type": "dummy", "key": DUMMYKEY}
        ]

        a._reg_queue.empty.side_effect = lambda: (len(queue) == 0)
        a._reg_queue.get.side_effect = lambda: queue.pop(0)

        def killloop(*args, **kwargs):
            a._running = False

        with mock.patch('gevent.sleep', side_effect=killloop):
            with mock.patch.object(a, '_send', side_effect=InvalidRequest) as send:
                a._process_queue()

                send.assert_not_called()
                a._mdns_updater.P2P_disable.assert_not_called()

    def test_process_queue_handles_exception_in_unqueueing(self):
        """An exception in unqueing an item should reset the object state to unregistered."""

        a = Aggregator(mdns_updater=mock.MagicMock())
        a._node_data["registered"] = True
        a.aggregator = "www.example.com"

        a._reg_queue.empty.return_value = False
        a._reg_queue.get.side_effect = Exception

        def killloop(*args, **kwargs):
            a._running = False

        with mock.patch('gevent.sleep', side_effect=killloop):
            with mock.patch.object(a, '_send') as send:
                a._process_queue()

                send.assert_not_called()
                a._mdns_updater.P2P_disable.assert_called_with()
                self.assertFalse(a._node_data["registered"])

    # # ================================================================================================================
    # # Test _send method
    # # ================================================================================================================
    def assert_send_runs_correctly(
        self, method,
        aggregator, api_ver,
        url, data=None,
        request=None,
        expected_return=None,
        expected_exception=None,
        prefer_ipv6=False
    ):
        """Function to test that the _send() method runs correctly"""

        def create_mock_request(method, aggregator_url, api_ver, url, expected_data, prefer_ipv6=False):
            if not prefer_ipv6:
                return (mock.call(
                    method=method,
                    url=urljoin(
                        aggregator_url,
                        AGGREGATOR_APINAMESPACE + "/" + AGGREGATOR_APINAME + "/" + a.aggregator_apiversion + '/' + url
                    ),
                    json=expected_data,
                    timeout=1.0))
            else:
                return (mock.call(
                    method=method,
                    url=urljoin(
                        aggregator_url,
                        AGGREGATOR_APINAMESPACE + "/" + AGGREGATOR_APINAME + "/" + a.aggregator_apiversion + '/' + url
                    ),
                    json=expected_data,
                    timeout=1.0,
                    proxies={'http': ''}))

        a = Aggregator(mdns_updater=mock.MagicMock())

        expected_request_calls = []
        expected_request_calls.append(
            create_mock_request(method, aggregator, api_ver, url, data, prefer_ipv6)
        )

        with mock.patch.dict(nmosnode.aggregator._config, {'prefer_ipv6': prefer_ipv6}):
            with mock.patch("requests.request", side_effect=request) as _request:
                R = None
                if expected_exception is not None:
                    with self.assertRaises(expected_exception):
                        R = a._send(method, aggregator, api_ver, url, data)
                else:
                    try:
                        R = a._send(method, aggregator, api_ver, url, data)
                    except Exception as e:
                        self.fail(msg="_send threw an unexpected exception, {}".format(e))

                self.assertListEqual(_request.mock_calls, expected_request_calls)
                if R:
                    self.assertEqual(R.content, expected_return)

    def test_send_200_response(self):
        TEST_CONTENT = "kasjhdlkhnjgsn"

        def request(*args, **kwargs):
            return mock.MagicMock(status_code=200, content=TEST_CONTENT)
        self.assert_send_runs_correctly("POST", "http://www.example.com:80", "v1.2", "test",
                                        request=request, expected_return=TEST_CONTENT)

    def test_send_200_response_ipv6(self):
        TEST_CONTENT = "kasjhdlkhnjgsn"

        def request(*args, **kwargs):
            return mock.MagicMock(status_code=200, content=TEST_CONTENT)
        self.assert_send_runs_correctly("POST", "http://www.example.com:80", "v1.2", "test",
                                        request=request, expected_return=TEST_CONTENT, prefer_ipv6=True)

    def test_send_201_response(self):
        TEST_CONTENT = "kasjhdlkhnjgsn"

        def request(*args, **kwargs):
            return mock.MagicMock(status_code=201, content=TEST_CONTENT)
        self.assert_send_runs_correctly("POST", "http://www.example.com:80", "v1.2", "test",
                                        request=request, expected_return=TEST_CONTENT)

    def test_send_204_response(self):
        TEST_CONTENT = "kasjhdlkhnjgsn"

        def request(*args, **kwargs):
            return mock.MagicMock(status_code=204, content=TEST_CONTENT)
        self.assert_send_runs_correctly("POST", "http://www.example.com:80", "v1.2", "test",
                                        request=request, expected_return=TEST_CONTENT)

    def test_send_409_response(self):
        TEST_CONTENT = "kasjhdlkhnjgsn"

        def request(*args, **kwargs):
            return mock.MagicMock(status_code=409, content=TEST_CONTENT)
        self.assert_send_runs_correctly("POST", "http://www.example.com:80", "v1.2", "test",
                                        request=request, expected_return=TEST_CONTENT)

    def test_send_4xx_response(self):
        """On a 4xx response an InvalidRequest Exception should be raised
        apart from 409, which is handled separately"""
        def request(*args, **kwargs):
            return mock.MagicMock(status_code=401)
        self.assert_send_runs_correctly("POST", "http://www.example.com:80", "v1.2", "test",
                                        request=request, expected_exception=InvalidRequest)

    def test_send_5xx_response(self):
        """On a 5xx response a ServerSideError Exception should be raised"""
        def request(*args, **kwargs):
            return mock.MagicMock(status_code=500)
        self.assert_send_runs_correctly("POST", "http://www.example.com:80", "v1.2", "test",
                                        request=request, expected_exception=ServerSideError)

    def test_send_no_response(self):
        """On a no response a ServerSideError Exception should be raised"""
        def request(*args, **kwargs):
            return None
        self.assert_send_runs_correctly("POST", "http://www.example.com:80", "v1.2", "test",
                                        request=request, expected_exception=ServerSideError)

    def test_send_request_exception(self):
        """On a request exception a ServerSideError Exception should be raised"""
        def request(*args, **kwargs):
            raise requests.exceptions.RequestException
        self.assert_send_runs_correctly("POST", "http://www.example.com:80", "v1.2", "test",
                                        request=request, expected_exception=ServerSideError)

    # ==================================================================================================================
    # Test main thread
    # ==================================================================================================================

    def test_main_thread_under_normal_operation(self):
        """Test the main thread under normal operation"""
        DUMMYNODEID = "90f7c2c0-cfa9-11e7-9b9d-2fe338e1e7ce"
        AGGREGATOR_1 = "http://example1.com"

        a = Aggregator(mdns_updater=mock.MagicMock())
        a._node_data["registered"] = False
        a._node_data["node"] = {"type": "node", "data": {"id": DUMMYNODEID}}

        def killloop(*args, **kwargs):
            if a._node_data['registered']:
                a._running = False

        request_mocks = [
            mock.MagicMock(name="request1()", status_code=404),
            mock.MagicMock(name="request2()", status_code=201),
            mock.MagicMock(name="request3()", status_code=200),
        ]

        expected_request_calls = [
            mock.call(method='POST', url=self.construct_url(AGGREGATOR_1, a.aggregator_apiversion,
                      "health/nodes/{}".format(DUMMYNODEID)), json=None, timeout=1.0),
            mock.call(method='POST', url=self.construct_url(AGGREGATOR_1, a.aggregator_apiversion, "resource"),
                      json={"data": {"id": DUMMYNODEID}, "type": "node"}, timeout=1.0),
            mock.call(method='POST', url=self.construct_url(AGGREGATOR_1, a.aggregator_apiversion,
                      "health/nodes/{}".format(DUMMYNODEID)), json=None, timeout=1.0),
        ]

        a.mdnsbridge.getHrefWithException.return_value = AGGREGATOR_1

        with mock.patch('gevent.sleep', side_effect=killloop):
            with mock.patch('requests.request', side_effect=request_mocks) as request:

                a._main_thread()
                self.assertTrue(a._node_data["registered"])
                self.assertListEqual(request.mock_calls, expected_request_calls)

    def test_main_thread_when_heartbeat_200(self):
        """Test the main thread when heartbeat returns 200 response"""
        DUMMYNODEID = "90f7c2c0-cfa9-11e7-9b9d-2fe338e1e7ce"
        AGGREGATOR_1 = "http://example1.com"

        a = Aggregator(mdns_updater=mock.MagicMock())
        a._node_data["registered"] = False
        a._node_data["node"] = {"type": "node", "data": {"id": DUMMYNODEID}}

        def killloop(*args, **kwargs):
            if a._node_data['registered']:
                a._running = False

        request_mocks = [
            mock.MagicMock(name="request1()", status_code=200, headers={}),
            mock.MagicMock(name="request2()", status_code=204),
            mock.MagicMock(name="request3()", status_code=201),
            mock.MagicMock(name="request4()", status_code=200),
        ]

        expected_request_calls = [
            mock.call(method='POST', url=self.construct_url(AGGREGATOR_1, a.aggregator_apiversion,
                      "health/nodes/{}".format(DUMMYNODEID)), json=None, timeout=1.0),
            mock.call(method='DELETE', url=self.construct_url(AGGREGATOR_1, a.aggregator_apiversion,
                      "resource/nodes/{}".format(DUMMYNODEID)), json=None, timeout=1.0),
            mock.call(method='POST', url=self.construct_url(AGGREGATOR_1, a.aggregator_apiversion, "resource"),
                      json={"data": {"id": DUMMYNODEID}, "type": "node"}, timeout=1.0),
            mock.call(method='POST', url=self.construct_url(AGGREGATOR_1, a.aggregator_apiversion,
                      "health/nodes/{}".format(DUMMYNODEID)), json=None, timeout=1.0),
        ]

        a.mdnsbridge.getHrefWithException.return_value = AGGREGATOR_1

        with mock.patch('gevent.sleep', side_effect=killloop):
            with mock.patch('requests.request', side_effect=request_mocks) as request:

                a._main_thread()
                self.assertTrue(a._node_data["registered"])
                self.assertListEqual(request.mock_calls, expected_request_calls)

    def test_main_thread_when_heartbeat_5xx(self):
        """Test the main thread when heartbeat returns 200 response"""
        DUMMYNODEID = "90f7c2c0-cfa9-11e7-9b9d-2fe338e1e7ce"
        AGGREGATOR_1 = "http://example1.com"
        AGGREGATOR_2 = "http://example2.com"

        a = Aggregator(mdns_updater=mock.MagicMock())
        a._node_data["registered"] = False
        a._node_data["node"] = {"type": "node", "data": {"id": DUMMYNODEID}}

        def killloop(*args, **kwargs):
            if a._node_data['registered']:
                a._running = False

        request_mocks = [
            mock.MagicMock(name="request1()", status_code=500),
            mock.MagicMock(name="request2()", status_code=404),
            mock.MagicMock(name="request3()", status_code=201),
            mock.MagicMock(name="request4()", status_code=200),
        ]

        expected_request_calls = [
            mock.call(method='POST', url=self.construct_url(AGGREGATOR_1, a.aggregator_apiversion,
                      "health/nodes/{}".format(DUMMYNODEID)), json=None, timeout=1.0),
            mock.call(method='POST', url=self.construct_url(AGGREGATOR_2, a.aggregator_apiversion,
                      "health/nodes/{}".format(DUMMYNODEID)), json=None, timeout=1.0),
            mock.call(method='POST', url=self.construct_url(AGGREGATOR_2, a.aggregator_apiversion, "resource"),
                      json={"data": {"id": DUMMYNODEID}, "type": "node"}, timeout=1.0),
            mock.call(method='POST', url=self.construct_url(AGGREGATOR_2, a.aggregator_apiversion,
                      "health/nodes/{}".format(DUMMYNODEID)), json=None, timeout=1.0),
        ]

        a.mdnsbridge.getHrefWithException.side_effect = [AGGREGATOR_1, AGGREGATOR_2]

        with mock.patch('gevent.sleep', side_effect=killloop):
            with mock.patch('requests.request', side_effect=request_mocks) as request:

                a._main_thread()
                self.assertTrue(a._node_data["registered"])
                self.assertListEqual(request.mock_calls, expected_request_calls)

    def test_main_thread_when_register_200(self):
        """Test the main thread when register returns 200 response"""
        DUMMYNODEID = "90f7c2c0-cfa9-11e7-9b9d-2fe338e1e7ce"
        AGGREGATOR_1 = "http://example1.com"

        a = Aggregator(mdns_updater=mock.MagicMock())
        a._node_data["registered"] = False
        a._node_data["node"] = {"type": "node", "data": {"id": DUMMYNODEID}}

        def killloop(*args, **kwargs):
            if a._node_data['registered']:
                a._running = False

        request_mocks = [
            mock.MagicMock(name="request1()", status_code=404),
            mock.MagicMock(name="request2()", status_code=200, headers={}),
            mock.MagicMock(name="request3()", status_code=204),
            mock.MagicMock(name="request4()", status_code=201),
            mock.MagicMock(name="request5()", status_code=200),
        ]

        expected_request_calls = [
            mock.call(method='POST', url=self.construct_url(AGGREGATOR_1, a.aggregator_apiversion,
                      "health/nodes/{}".format(DUMMYNODEID)), json=None, timeout=1.0),
            mock.call(method='POST', url=self.construct_url(AGGREGATOR_1, a.aggregator_apiversion, "resource"),
                      json={"data": {"id": DUMMYNODEID}, "type": "node"}, timeout=1.0),
            mock.call(method='DELETE', url=self.construct_url(AGGREGATOR_1, a.aggregator_apiversion,
                      "resource/nodes/{}".format(DUMMYNODEID)), json=None, timeout=1.0),
            mock.call(method='POST', url=self.construct_url(AGGREGATOR_1, a.aggregator_apiversion, "resource"),
                      json={"data": {"id": DUMMYNODEID}, "type": "node"}, timeout=1.0),
            mock.call(method='POST', url=self.construct_url(AGGREGATOR_1, a.aggregator_apiversion,
                      "health/nodes/{}".format(DUMMYNODEID)), json=None, timeout=1.0),
        ]

        a.mdnsbridge.getHrefWithException.return_value = AGGREGATOR_1

        with mock.patch('gevent.sleep', side_effect=killloop):
            with mock.patch('requests.request', side_effect=request_mocks) as request:

                a._main_thread()
                self.assertTrue(a._node_data["registered"])
                self.assertListEqual(request.mock_calls, expected_request_calls)

    def test_main_thread_when_register_5xx(self):
        """Test the main thread when heartbeat returns 200 response"""
        DUMMYNODEID = "90f7c2c0-cfa9-11e7-9b9d-2fe338e1e7ce"
        AGGREGATOR_1 = "http://example1.com"
        AGGREGATOR_2 = "http://example2.com"

        a = Aggregator(mdns_updater=mock.MagicMock())
        a._node_data["registered"] = False
        a._node_data["node"] = {"type": "node", "data": {"id": DUMMYNODEID}}

        def killloop(*args, **kwargs):
            if a._node_data['registered']:
                a._running = False

        request_mocks = [
            mock.MagicMock(name="request1()", status_code=404),
            mock.MagicMock(name="request2()", status_code=500),
            mock.MagicMock(name="request3()", status_code=404),
            mock.MagicMock(name="request4()", status_code=201),
            mock.MagicMock(name="request5()", status_code=200),
        ]

        expected_request_calls = [
            mock.call(method='POST', url=self.construct_url(AGGREGATOR_1, a.aggregator_apiversion,
                      "health/nodes/{}".format(DUMMYNODEID)), json=None, timeout=1.0),
            mock.call(method='POST', url=self.construct_url(AGGREGATOR_1, a.aggregator_apiversion, "resource"),
                      json={"data": {"id": DUMMYNODEID}, "type": "node"}, timeout=1.0),
            mock.call(method='POST', url=self.construct_url(AGGREGATOR_2, a.aggregator_apiversion,
                      "health/nodes/{}".format(DUMMYNODEID)), json=None, timeout=1.0),
            mock.call(method='POST', url=self.construct_url(AGGREGATOR_2, a.aggregator_apiversion, "resource"),
                      json={"data": {"id": DUMMYNODEID}, "type": "node"}, timeout=1.0),
            mock.call(method='POST', url=self.construct_url(AGGREGATOR_2, a.aggregator_apiversion,
                      "health/nodes/{}".format(DUMMYNODEID)), json=None, timeout=1.0),
        ]

        a.mdnsbridge.getHrefWithException.side_effect = [AGGREGATOR_1, AGGREGATOR_2]

        with mock.patch('gevent.sleep', side_effect=killloop):
            with mock.patch('requests.request', side_effect=request_mocks) as request:

                a._main_thread()
                self.assertTrue(a._node_data["registered"])
                self.assertListEqual(request.mock_calls, expected_request_calls)



# Copyright 2019 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from nmoscommon.ipc import Host
from nmoscommon.logger import Logger

ADDRESS = "ipc:///tmp/ips-nodefacade"


def ipcmethod(name=None):
    def decorator(function):
        function.ipc_method = True
        function.ipc_name = name
        return function
    if callable(name):
        return decorator(name)
    return decorator


class FacadeInterface(object):
    def __init__(self, registry, logger):
        self.host = Host(ADDRESS)
        self.registry = registry
        self.logger = Logger("facade_interface", logger)

        def getbases(cl):
            bases = list(cl.__bases__)
            for x in cl.__bases__:
                bases += getbases(x)
            return bases

        for cl in [self.__class__, ] + getbases(self.__class__):
            for name in cl.__dict__.keys():
                value = getattr(self, name)
                if callable(value):
                    if hasattr(value, "ipc_method"):
                        self.host.ipcmethod(name)(value)

    def start(self):
        self.host.start()

    def stop(self):
        self.host.stop()

    @ipcmethod
    def srv_register(self, name, srv_type, pid, href, proxy_path, authorization=False):
        self.logger.writeInfo("Service Register {}, {}, {}, {}, {}".format(name, srv_type, pid, href, proxy_path))
        return self.registry.register_service(name, srv_type, pid, href, proxy_path, authorization)

    # TODO: =None should be removed once proxying removed from node facade
    @ipcmethod
    def srv_update(self, name, pid, href, proxy_path):
        self.logger.writeInfo("Service Update {}, {}, {}, {}".format(name, pid, href, proxy_path))
        return self.registry.update_service(name, pid, href, proxy_path)

    @ipcmethod
    def srv_unregister(self, name, pid):
        self.logger.writeInfo("Service Unregister {}, {}".format(name, pid))
        return self.registry.unregister_service(name, pid)

    @ipcmethod
    def srv_heartbeat(self, name, pid):
        self.logger.writeDebug("Service Heartbeat {}, {}".format(name, pid))
        return self.registry.heartbeat_service(name, pid)

    @ipcmethod
    def res_register(self, name, pid, type, key, value):
        self.logger.writeInfo("Resource Register {} {} {} {} {}".format(name, pid, type, key, value))
        return self.registry.register_resource(name, pid, type, key, value)

    @ipcmethod
    def res_update(self, name, pid, type, key, value):
        self.logger.writeInfo("Resource Update {} {} {} {} {}".format(name, pid, type, key, value))
        return self.registry.update_resource(name, pid, type, key, value)

    @ipcmethod
    def res_unregister(self, name, pid, type, key):
        self.logger.writeInfo("Resource Unregister {} {} {} {}".format(name, pid, type, key))
        return self.registry.unregister_resource(name, pid, type, key)

    @ipcmethod
    def control_register(self, name, pid, device_id, control_data):
        self.logger.writeInfo("Control Register {} {} {} {}".format(name, pid, device_id, control_data))
        return self.registry.register_control(name, pid, device_id, control_data)

    @ipcmethod
    def control_unregister(self, name, pid, device_id, control_data):
        self.logger.writeInfo("Control Unregister {} {} {} {}".format(name, pid, device_id, control_data))
        return self.registry.unregister_control(name, pid, device_id, control_data)

    @ipcmethod
    def self_get(self, name, pid, api_version):
        return self.registry.list_self(api_version)

    @ipcmethod
    def status_get(self, name, pid):
        return self.registry.aggregator.status()

    @ipcmethod
    def clock_register(self, name, pid, clk_data):
        self.logger.writeInfo("Clock Register {} {}".format(name, pid))
        return self.registry.register_clock(clk_data)

    @ipcmethod
    def clock_update(self, name, pid, clk_data):
        self.logger.writeInfo("Clock Update {} {}".format(name, pid))
        return self.registry.update_clock(clk_data)

    @ipcmethod
    def clock_unregister(self, name, pid, clk_name):
        self.logger.writeInfo("Clock Unregister {} {}".format(name, pid))
        return self.registry.unregister_clock(clk_name)



# Copyright 2019 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import print_function
from six import PY2
from six import iteritems
import unittest
import mock
from nmosnode.facade import Facade, FAC_SUCCESS, FAC_OTHERERROR


class TestFacade(unittest.TestCase):
    def setUp(self):
        paths = ['nmosnode.facade.Logger',
                 'nmosnode.facade.Proxy', ]
        patchers = {name: mock.patch(name) for name in paths}
        self.mocks = {name: patcher.start() for (name, patcher) in iteritems(patchers)}
        for (name, patcher) in iteritems(patchers):
            self.addCleanup(patcher.stop)

        def _invoke_named(obj):
            def __inner(method, *args, **kwargs):
                return getattr(obj, method)(*args, **kwargs)
            return __inner
        self.mocks['nmosnode.facade.Proxy'].return_value.invoke_named.side_effect = _invoke_named(self.mocks['nmosnode.facade.Proxy'].return_value)

        def printmsg(t):
            def _inner(msg):
                print(t + ": " + msg)
            return _inner

#        self.mocks['nmosnode.facade.Logger'].return_value.writeInfo.side_effect = printmsg("INFO")
#        self.mocks['nmosnode.facade.Logger'].return_value.writeWarning.side_effect = printmsg("WARNING")
#        self.mocks['nmosnode.facade.Logger'].return_value.writeDebug.side_effect = printmsg("DEBUG")
#        self.mocks['nmosnode.facade.Logger'].return_value.writeError.side_effect = printmsg("ERROR")
#        self.mocks['nmosnode.facade.Logger'].return_value.writeFatal.side_effect = printmsg("FATAL")

        if PY2:
            self.assertCountEqual = self.assertItemsEqual

    def test_setup_ipc(self):
        address = "ipc:///tmp/nmos-nodefacade.dummy.for.test"
        UUT = Facade("dummy_type", address=address)
        UUT.setup_ipc()

        self.mocks['nmosnode.facade.Proxy'].assert_called_once_with(address)
        self.assertEqual(self.mocks['nmosnode.facade.Proxy'].return_value, UUT.ipc)

    def test_setup_ipc_exception_handling(self):
        address = "ipc:///tmp/nmos-nodefacade.dummy.for.test"
        UUT = Facade("dummy_type", address=address)

        self.mocks['nmosnode.facade.Proxy'].side_effect = Exception

        UUT.setup_ipc()

        self.mocks['nmosnode.facade.Proxy'].assert_called_once_with(address)
        self.assertIsNone(UUT.ipc)

    def test_register_service(self):
        address = "ipc:///tmp/nmos-nodefacade.dummy.for.test"
        srv_type = "dummy_type"
        UUT = Facade(srv_type, address=address)

        self.mocks['nmosnode.facade.Proxy'].return_value.srv_register.return_value = FAC_SUCCESS

        href = "http://dummy.example.com"
        proxy_path = "http://dummyproxy.example.com"
        UUT.register_service(href, proxy_path)

        self.assertIsNotNone(UUT.ipc)
        UUT.ipc.srv_register.assert_called_once_with(
            srv_type, "urn:x-ipstudio:service:" + srv_type, mock.ANY, href, proxy_path, False
        )
        self.assertTrue(UUT.srv_registered)

    def test_register_service_bails_when_no_ipc(self):
        address = "ipc:///tmp/nmos-nodefacade.dummy.for.test"
        srv_type = "dummy_type"
        UUT = Facade(srv_type, address=address)

        self.mocks['nmosnode.facade.Proxy'].side_effect = Exception

        href = "http://dummy.example.com"
        proxy_path = "http://dummyproxy.example.com"
        UUT.register_service(href, proxy_path)

        self.assertIsNone(UUT.ipc)
        self.assertFalse(UUT.srv_registered)

    def test_register_service_bails_when_register_fails(self):
        address = "ipc:///tmp/nmos-nodefacade.dummy.for.test"
        srv_type = "dummy_type"
        UUT = Facade(srv_type, address=address)

        self.mocks['nmosnode.facade.Proxy'].return_value.srv_register.return_value = FAC_OTHERERROR

        href = "http://dummy.example.com"
        proxy_path = "http://dummyproxy.example.com"
        UUT.register_service(href, proxy_path)

        self.assertIsNotNone(UUT.ipc)
        UUT.ipc.srv_register.assert_called_once_with(
            srv_type, "urn:x-ipstudio:service:" + srv_type, mock.ANY, href, proxy_path, False
        )
        self.assertFalse(UUT.srv_registered)

    def test_register_service_bails_when_register_raises(self):
        address = "ipc:///tmp/nmos-nodefacade.dummy.for.test"
        srv_type = "dummy_type"
        UUT = Facade(srv_type, address=address)

        self.mocks['nmosnode.facade.Proxy'].return_value.srv_register.side_effect = Exception

        href = "http://dummy.example.com"
        proxy_path = "http://dummyproxy.example.com"
        UUT.register_service(href, proxy_path)

        self.assertIsNone(UUT.ipc)
        self.mocks['nmosnode.facade.Proxy'].return_value.srv_register.assert_called_once_with(
            srv_type, "urn:x-ipstudio:service:" + srv_type, mock.ANY, href, proxy_path, False
        )
        self.assertFalse(UUT.srv_registered)

    def test_unregister_service(self):
        address = "ipc:///tmp/nmos-nodefacade.dummy.for.test"
        srv_type = "dummy_type"
        UUT = Facade(srv_type, address=address)

        UUT.srv_registered = True
        UUT.unregister_service()

        self.assertIsNotNone(UUT.ipc)
        UUT.ipc.srv_unregister.assert_called_once_with(srv_type, mock.ANY)
        self.assertFalse(UUT.srv_registered)

    def test_unregister_service_bails_when_no_ipc(self):
        address = "ipc:///tmp/nmos-nodefacade.dummy.for.test"
        srv_type = "dummy_type"
        UUT = Facade(srv_type, address=address)

        self.mocks['nmosnode.facade.Proxy'].side_effect = Exception

        UUT.srv_registered = True
        UUT.unregister_service()

        self.assertIsNone(UUT.ipc)
        self.assertTrue(UUT.srv_registered)

    def test_unregister_service_bails_when_unregister_raises(self):
        address = "ipc:///tmp/nmos-nodefacade.dummy.for.test"
        srv_type = "dummy_type"
        UUT = Facade(srv_type, address=address)

        self.mocks['nmosnode.facade.Proxy'].return_value.srv_unregister.side_effect = Exception

        UUT.srv_registered = True
        UUT.unregister_service()

        self.assertIsNone(UUT.ipc)
        self.mocks['nmosnode.facade.Proxy'].return_value.srv_unregister.assert_called_once_with(srv_type, mock.ANY)
        self.assertTrue(UUT.srv_registered)

    def test_heartbeat_service(self):
        address = "ipc:///tmp/nmos-nodefacade.dummy.for.test"
        srv_type = "dummy_type"
        UUT = Facade(srv_type, address=address)

        self.mocks['nmosnode.facade.Proxy'].return_value.srv_heartbeat.return_value = FAC_SUCCESS

        with mock.patch.object(UUT, 'reregister_all') as reregister_all:
            UUT.heartbeat_service()

            self.assertTrue(UUT.srv_registered)
            self.mocks['nmosnode.facade.Proxy'].return_value.srv_heartbeat.assert_called_once_with(srv_type, mock.ANY)
            reregister_all.assert_not_called()

    def test_heartbeat_service_bails_when_no_ipc(self):
        address = "ipc:///tmp/nmos-nodefacade.dummy.for.test"
        srv_type = "dummy_type"
        UUT = Facade(srv_type, address=address)

        self.mocks['nmosnode.facade.Proxy'].side_effect = Exception

        with mock.patch.object(UUT, 'reregister_all') as reregister_all:
            UUT.heartbeat_service()

            self.assertFalse(UUT.srv_registered)
            self.mocks['nmosnode.facade.Proxy'].return_value.srv_heartbeat.assert_not_called()
            reregister_all.assert_not_called()

    def test_heartbeat_service_bails_when_heartbeat_raises(self):
        address="ipc:///tmp/nmos-nodefacade.dummy.for.test"
        srv_type = "dummy_type"
        UUT = Facade(srv_type, address=address)

        self.mocks['nmosnode.facade.Proxy'].return_value.srv_heartbeat.side_effect = Exception

        with mock.patch.object(UUT, 'reregister_all') as reregister_all:
            UUT.heartbeat_service()

            self.assertFalse(UUT.srv_registered)
            self.mocks['nmosnode.facade.Proxy'].return_value.srv_heartbeat.assert_called_once_with(srv_type, mock.ANY)
            reregister_all.assert_not_called()

    def test_heartbeat_service_reregisters_on_failure(self):
        address="ipc:///tmp/nmos-nodefacade.dummy.for.test"
        srv_type = "dummy_type"
        UUT = Facade(srv_type, address=address)

        self.mocks['nmosnode.facade.Proxy'].return_value.srv_heartbeat.return_value = FAC_OTHERERROR

        with mock.patch.object(UUT, 'reregister_all') as reregister_all:
            UUT.heartbeat_service()

            self.assertFalse(UUT.srv_registered)
            self.mocks['nmosnode.facade.Proxy'].return_value.srv_heartbeat.assert_called_once_with(srv_type, mock.ANY)
            reregister_all.assert_called_once_with()

    def test_heartbeat_service_reregisters_on_success_when_told_to(self):
        """Is this correct? This is current behaviour, but doesn't match the comments"""
        address = "ipc:///tmp/nmos-nodefacade.dummy.for.test"
        srv_type = "dummy_type"
        UUT = Facade(srv_type, address=address)

        self.mocks['nmosnode.facade.Proxy'].return_value.srv_heartbeat.return_value = FAC_SUCCESS

        with mock.patch.object(UUT, 'reregister_all') as reregister_all:
            UUT.reregister = True
            UUT.heartbeat_service()

            self.assertTrue(UUT.srv_registered)
            self.mocks['nmosnode.facade.Proxy'].return_value.srv_heartbeat.assert_called_once_with(srv_type, mock.ANY)
            reregister_all.assert_called_once_with()

    def assert_method_calls_remote_method_or_bails(self, method_name, remote_method_name, params, registered=True, ipc=True, raises=False, presetup=None, extra_check=None):
        getattr(self.mocks['nmosnode.facade.Proxy'].return_value, remote_method_name).reset_mock()
        getattr(self.mocks['nmosnode.facade.Proxy'].return_value, remote_method_name).side_effect = None
        self.mocks['nmosnode.facade.Proxy'].side_effect = None

        address = "ipc:///tmp/nmos-nodefacade.dummy.for.test"
        srv_type = "dummy_type"
        UUT = Facade(srv_type, address=address)

        if presetup is not None:
            presetup(UUT)

        self.mocks['nmosnode.facade.Proxy'].return_value.srv_register.return_value = FAC_SUCCESS

        if registered:
            href = "http://dummy.example.com"
            proxy_path = "http://dummyproxy.example.com"
            UUT.register_service(href, proxy_path)

        if not ipc:
            UUT.ipc = None
            self.mocks['nmosnode.facade.Proxy'].side_effect = Exception

        if raises:
            getattr(self.mocks['nmosnode.facade.Proxy'].return_value, remote_method_name).side_effect = Exception

        getattr(UUT, method_name)(*params)
        if registered and ipc:
            callparams = [srv_type, mock.ANY] + list(params)
            getattr(self.mocks['nmosnode.facade.Proxy'].return_value, remote_method_name).assert_called_once_with(*callparams)
        elif not registered or not ipc:
            getattr(self.mocks['nmosnode.facade.Proxy'].return_value, remote_method_name).assert_not_called()
            self.assertTrue(UUT.reregister)
        if raises:
            self.assertTrue(UUT.reregister)
        if extra_check is not None:
            extra_check(UUT)

    def test_addResource(self):
        def _extra_check(UUT):
            self.assertIn("dummykey", UUT.resources["dummytype"])
            self.assertEqual(UUT.resources["dummytype"]["dummykey"], "dummyval")
        self.assert_method_calls_remote_method_or_bails('addResource', 'res_register', ("dummytype", "dummykey", "dummyval"), extra_check=_extra_check)
        self.assert_method_calls_remote_method_or_bails('addResource', 'res_register', ("dummytype", "dummykey", "dummyval"), registered=False)
        self.assert_method_calls_remote_method_or_bails('addResource', 'res_register', ("dummytype", "dummykey", "dummyval"), ipc=False)
        self.assert_method_calls_remote_method_or_bails('addResource', 'res_register', ("dummytype", "dummykey", "dummyval"), raises=True)

    def test_updateResource(self):
        def _extra_check(UUT):
            self.assertIn("dummykey", UUT.resources["dummytype"])
            self.assertEqual(UUT.resources["dummytype"]["dummykey"], "dummyval")
        self.assert_method_calls_remote_method_or_bails('updateResource', 'res_update', ("dummytype", "dummykey", "dummyval"), extra_check=_extra_check)
        self.assert_method_calls_remote_method_or_bails('updateResource', 'res_update', ("dummytype", "dummykey", "dummyval"), registered=False)
        self.assert_method_calls_remote_method_or_bails('updateResource', 'res_update', ("dummytype", "dummykey", "dummyval"), ipc=False)
        self.assert_method_calls_remote_method_or_bails('updateResource', 'res_update', ("dummytype", "dummykey", "dummyval"), raises=True)

    def test_delResource(self):
        def _presetup(UUT):
            if "flow" not in UUT.resources:
                UUT.resources["flow"] = {}
            if "transport" not in UUT.resources:
                UUT.resources["transport"] = {}
            UUT.resources["flow"]["dummykey"] = 'dummyval'
            UUT.resources["transport"]["dummytransportkey"] = { "flow-id" : "dummykey" }

        def _extra_check(UUT):
            self.assertNotIn("dummykey", UUT.resources["flow"])
            self.assertNotIn("dummytransportkey", UUT.resources["transport"])
        self.assert_method_calls_remote_method_or_bails('delResource', 'res_unregister', ("flow", "dummykey"), presetup=_presetup, extra_check=_extra_check)
        self.assert_method_calls_remote_method_or_bails('delResource', 'res_unregister', ("dummytype", "dummykey"), registered=False)
        self.assert_method_calls_remote_method_or_bails('delResource', 'res_unregister', ("dummytype", "dummykey"), ipc=False)
        self.assert_method_calls_remote_method_or_bails('delResource', 'res_unregister', ("dummytype", "dummykey"), raises=True)

    def test_addControl(self):
        def _extra_check(UUT):
            self.assertIn("dummyid", UUT.controls)
            self.assertIn("dummyhref", UUT.controls["dummyid"])
            self.assertEqual(UUT.controls["dummyid"]["dummyhref"], { "href" : "dummyhref" })
        self.assert_method_calls_remote_method_or_bails('addControl', 'control_register', ("dummyid", {"href": "dummyhref"}), extra_check=_extra_check)
        self.assert_method_calls_remote_method_or_bails('addControl', 'control_register', ("dummyid", {"href": "dummyhref"}), registered=False)
        self.assert_method_calls_remote_method_or_bails('addControl', 'control_register', ("dummyid", {"href": "dummyhref"}), ipc=False)
        self.assert_method_calls_remote_method_or_bails('addControl', 'control_register', ("dummyid", {"href": "dummyhref"}), raises=True)

    def test_delControl(self):
        def _presetup(UUT):
            if "dummyid" not in UUT.controls:
                UUT.controls["dummyid"] = {}
            UUT.controls["dummyid"]["dummyhref"] = 'dummyval'
        def _extra_check(UUT):
            self.assertNotIn("dummyhref", UUT.controls["dummyid"])
        self.assert_method_calls_remote_method_or_bails('delControl', 'control_unregister', ("dummyid", {"href": "dummyhref"}), presetup=_presetup, extra_check=_extra_check)
        self.assert_method_calls_remote_method_or_bails('delControl', 'control_unregister', ("dummyid", {"href": "dummyhref"}), registered=False)
        self.assert_method_calls_remote_method_or_bails('delControl', 'control_unregister', ("dummyid", {"href": "dummyhref"}), ipc=False)
        self.assert_method_calls_remote_method_or_bails('delControl', 'control_unregister', ("dummyid", {"href": "dummyhref"}), raises=True)

    def test_get_node_self(self):
        self.assert_method_calls_remote_method_or_bails('get_node_self', 'self_get', ("v1.1",))
        self.assert_method_calls_remote_method_or_bails('get_node_self', 'self_get', ("v1.1",), registered=False)
        self.assert_method_calls_remote_method_or_bails('get_node_self', 'self_get', ("v1.1",), ipc=False)
        self.assert_method_calls_remote_method_or_bails('get_node_self', 'self_get', ("v1.1",), raises=True)

    def test_debug_message(self):
        """There's not a lot we can sensibly check here, but we might as well check that every error has a message
        and that no two errors have the same message"""
        codes = [name for name in dir() if name.startswith('FAC_') ]
        address = "ipc:///tmp/nmos-nodefacade.dummy.for.test"
        srv_type = "dummy_type"
        UUT = Facade(srv_type, address=address)

        msgs = [UUT.debug_message(c) for c in codes]

        for msg in msgs:
            self.assertEqual(msgs.count(msg), 1)

    def test_reregister_all(self):
        address = "ipc:///tmp/nmos-nodefacade.dummy.for.test"
        srv_type = "dummy_type"
        UUT = Facade(srv_type, address=address)

        self.mocks['nmosnode.facade.Proxy'].return_value.srv_register.return_value = FAC_SUCCESS

        href = "http://dummy.example.com"
        proxy_path = "http://dummyproxy.example.com"
        UUT.register_service(href, proxy_path)

        resources = [("type0", "key0", "val0"),
                     ("type0", "key1", "val1"),
                     ("type2", "key2", "val2")]
        controls = [("id0", {"href": "href0"}),
                    ("id1", {"href": "href1"}),
                    ("id2", {"href": "href2"})]

        for res in resources:
            UUT.addResource(*res)
        UUT.addResource("receiver", "rkey", {"pipel_id": "DUMMY0", "pipeline_id": "DUMMY1", "dummy": "DUMMY2"})

        for con in controls:
            UUT.addControl(*con)

        expected_res_register_calls = [mock.call(srv_type, mock.ANY, *res) for res in resources] + [mock.call(srv_type, mock.ANY, "receiver", "rkey", {"dummy": "DUMMY2"})]
        expected_control_register_calls = [mock.call(srv_type, mock.ANY, *con) for con in controls]

        UUT.ipc.res_register.reset_mock()
        UUT.ipc.control_register.reset_mock()

        UUT.reregister_all()

        self.assertFalse(UUT.reregister)
        self.assertTrue(UUT.srv_registered)
        self.assertCountEqual(UUT.ipc.res_register.mock_calls, expected_res_register_calls)
        self.assertCountEqual(UUT.ipc.control_register.mock_calls, expected_control_register_calls)

    def test_reregister_all_bails_if_failed_to_unregister(self):
        address = "ipc:///tmp/nmos-nodefacade.dummy.for.test"
        srv_type = "dummy_type"
        UUT = Facade(srv_type, address=address)

        self.mocks['nmosnode.facade.Proxy'].return_value.srv_register.return_value = FAC_SUCCESS
        self.mocks['nmosnode.facade.Proxy'].return_value.srv_unregister.side_effect = Exception

        href = "http://dummy.example.com"
        proxy_path = "http://dummyproxy.example.com"
        UUT.register_service(href, proxy_path)

        UUT.ipc.res_register.reset_mock()
        UUT.ipc.control_register.reset_mock()

        UUT.reregister_all()

        self.mocks['nmosnode.facade.Proxy'].return_value.res_register.assert_not_called()
        self.mocks['nmosnode.facade.Proxy'].return_value.control_register.assert_not_called()

    def test_reregister_all_bails_if_failed_to_register(self):
        address = "ipc:///tmp/nmos-nodefacade.dummy.for.test"
        srv_type = "dummy_type"
        UUT = Facade(srv_type, address=address)

        self.mocks['nmosnode.facade.Proxy'].return_value.srv_register.return_value = FAC_SUCCESS

        href = "http://dummy.example.com"
        proxy_path = "http://dummyproxy.example.com"
        UUT.register_service(href, proxy_path)

        UUT.ipc.res_register.reset_mock()
        UUT.ipc.control_register.reset_mock()

        self.mocks['nmosnode.facade.Proxy'].return_value.srv_register.side_effect = Exception

        UUT.reregister_all()

        self.mocks['nmosnode.facade.Proxy'].return_value.res_register.assert_not_called()
        self.mocks['nmosnode.facade.Proxy'].return_value.control_register.assert_not_called()

    def test_reregister_all_bails_when_res_register_raises(self):
        address = "ipc:///tmp/nmos-nodefacade.dummy.for.test"
        srv_type = "dummy_type"
        UUT = Facade(srv_type, address=address)

        self.mocks['nmosnode.facade.Proxy'].return_value.srv_register.return_value = FAC_SUCCESS

        href = "http://dummy.example.com"
        proxy_path = "http://dummyproxy.example.com"
        UUT.register_service(href, proxy_path)

        resources = [("type0", "key0", "val0"),
                     ("type0", "key1", "val1"),
                     ("type2", "key2", "val2")]
        controls = [("id0", {"href": "href0"}),
                    ("id1", {"href": "href1"}),
                    ("id2", {"href": "href2"})]

        for res in resources:
            UUT.addResource(*res)
        UUT.addResource("receiver", "rkey", {"pipel_id": "DUMMY0", "pipeline_id": "DUMMY1", "dummy": "DUMMY2"})

        for con in controls:
            UUT.addControl(*con)

        UUT.ipc.res_register.reset_mock()
        UUT.ipc.control_register.reset_mock()

        UUT.ipc.res_register.side_effect = Exception

        with mock.patch('gevent.sleep'):
            UUT.reregister_all()

        self.assertIsNone(UUT.ipc)
        self.assertEqual(len(self.mocks['nmosnode.facade.Proxy'].return_value.res_register.mock_calls), 1)
        self.assertEqual(len(self.mocks['nmosnode.facade.Proxy'].return_value.control_register.mock_calls), 0)

    def test_reregister_all_bails_when_control_register_raises(self):
        address = "ipc:///tmp/nmos-nodefacade.dummy.for.test"
        srv_type = "dummy_type"
        UUT = Facade(srv_type, address=address)

        self.mocks['nmosnode.facade.Proxy'].return_value.srv_register.return_value = FAC_SUCCESS

        href = "http://dummy.example.com"
        proxy_path = "http://dummyproxy.example.com"
        UUT.register_service(href, proxy_path)

        resources = [("type0", "key0", "val0"),
                     ("type0", "key1", "val1"),
                     ("type2", "key2", "val2")]
        controls = [("id0", {"href": "href0"}),
                    ("id1", {"href": "href1"}),
                    ("id2", {"href": "href2"})]

        for res in resources:
            UUT.addResource(*res)
        UUT.addResource("receiver", "rkey", {"pipel_id": "DUMMY0", "pipeline_id": "DUMMY1", "dummy": "DUMMY2"})

        expected_res_register_calls = [mock.call(srv_type, mock.ANY, *res) for res in resources] + [mock.call(srv_type, mock.ANY, "receiver", "rkey", {"dummy": "DUMMY2"})]

        for con in controls:
            UUT.addControl(*con)

        UUT.ipc.res_register.reset_mock()
        UUT.ipc.control_register.reset_mock()

        UUT.ipc.control_register.side_effect = Exception

        with mock.patch('gevent.sleep'):
            UUT.reregister_all()

        self.assertIsNone(UUT.ipc)
        self.assertCountEqual(self.mocks['nmosnode.facade.Proxy'].return_value.res_register.mock_calls, expected_res_register_calls)
        self.assertEqual(len(self.mocks['nmosnode.facade.Proxy'].return_value.control_register.mock_calls), 1)



# Copyright 2019 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

import requests

from os import urandom
from flask import request, url_for, redirect
from socket import gethostname
from six import itervalues
from six.moves.urllib.parse import urljoin

from nmoscommon.nmoscommonconfig import config as _config
from nmoscommon.webapi import WebAPI, route, resource_route, abort

NODE_APIVERSIONS = ["v1.0", "v1.1", "v1.2", "v1.3"]
if _config.get("https_mode", "disabled") == "enabled":
    NODE_APIVERSIONS.remove("v1.0")

NODE_REGVERSION = _config.get('nodefacade', {}).get('NODE_REGVERSION', 'v1.2')

NODE_APINAMESPACE = "x-nmos"
NODE_APINAME = "node"
NODE_APIROOT = '/' + NODE_APINAMESPACE + '/' + NODE_APINAME + '/'
HOSTNAME = gethostname().split(".", 1)[0]
RESOURCE_TYPES = ["sources", "flows", "devices", "senders", "receivers"]


class FacadeAPI(WebAPI):
    def __init__(self, nmos_registry, auth_registry=None):
        super(FacadeAPI, self).__init__()
        self.app.config["SECRET_KEY"] = urandom(16)  # Required for 'session' in auth client
        self.registry = nmos_registry
        self.node_id = nmos_registry.node_id
        self.auth_registry = auth_registry
        self.auth_client = None
        if self.auth_registry:
            self.auth_registry.init_app(self.app)

    @route('/')
    def root(self):
        return [NODE_APINAMESPACE + "/"]

    @route('/' + NODE_APINAMESPACE + '/')
    def namespaceroot(self):
        return [NODE_APINAME + "/"]

    @route(NODE_APIROOT)
    def nameroot(self):
        return [api_version + "/" for api_version in NODE_APIVERSIONS]

    @route(NODE_APIROOT + "oauth/", auto_json=False)
    def oauth(self):
        """Redirect to Auth Server's authorization endpoint"""
        redirect_uri = url_for('_authorization', _external=True)
        if self.auth_registry.client_name:
            self.auth_client = getattr(self.auth_registry, self.auth_registry.client_name)
            return self.auth_client.authorize_redirect(redirect_uri)
        else:
            abort(400, "Client not registered with Auth Server")

    @route(NODE_APIROOT + "authorize", auto_json=False)
    def authorization(self):
        """Authorize Auth Server redirect to obtain token then store in memory"""
        self.auth_client = getattr(self.auth_registry, self.auth_registry.client_name)
        token = self.auth_client.authorize_access_token()
        self.auth_registry.update_local_token(token)
        return redirect(url_for('_nameroot'))

    @route(NODE_APIROOT + "<api_version>/")
    def versionroot(self, api_version):
        if api_version not in NODE_APIVERSIONS:
            abort(404)
        return ["self/", "sources/", "flows/", "devices/", "senders/", "receivers/"]

    @resource_route(NODE_APIROOT + "<api_version>/<resource_type>/")
    def resource_list(self, api_version, resource_type):
        if api_version not in NODE_APIVERSIONS:
            abort(404)
        if resource_type == "self":
            return self.registry.list_self(api_version=api_version)
        elif resource_type not in RESOURCE_TYPES:
            abort(404)
        return list(itervalues(self.registry.list_resource(resource_type.rstrip("s"), api_version=api_version)))

    @resource_route(NODE_APIROOT + "<api_version>/<resource_type>/<resource_id>/")
    def resource_id(self, api_version, resource_type, resource_id):
        if api_version not in NODE_APIVERSIONS:
            abort(404)
        if resource_type not in RESOURCE_TYPES:
            abort(404)
        resources = self.registry.list_resource(resource_type.rstrip("s"), api_version=api_version)
        if resource_id in resources:
            return resources[resource_id]
        else:
            abort(404)

    @resource_route(NODE_APIROOT + "<api_version>/receivers/<receiver_id>/target",
                    methods=['PUT'])
    def receiver_id_subscription(self, api_version, receiver_id):
        if api_version not in NODE_APIVERSIONS:
            abort(404)

        receiver_service = self.registry.find_service("receiver", receiver_id)
        if receiver_service is None:
            abort(404)

        receiver_service_href = self.registry.get_service_href(receiver_service)

        if receiver_service_href is None:
            # Service doesn't specify an href
            return {}
        if str(receiver_service_href).isdigit():
            # Service doesn't exist
            abort(404)
        receiver_subs_href = "receivers/" + receiver_id + "/target"
        href = urljoin(receiver_service_href, receiver_subs_href) + "/"
        # TODO Handle all request types
        # TODO Move into proxy class?

        headers = dict(request.headers)
        headers['Accept'] = 'application/json'
        del headers['Host']

        print("Sending {} request to '{}' with headers={} and data='{}'".format(request.method, href,
                                                                                headers, request.data))

        try:
            resp = requests.request(request.method, href, params=request.args, data=request.data, headers=headers,
                                    allow_redirects=True, timeout=30)
        except Exception:
            abort(500)

        if not resp:
            abort(503)

        print(resp)

        if resp.status_code // 100 != 2:
            abort(resp.status_code)

        data = {}
        if len(resp.text) > 0:
            data = resp.json()
        else:
            return (204, '')
        if resp.status_code == 200:
            return (data)
        else:
            return (resp.status_code, data)



# Copyright 2019 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
import gevent


class MockBackend:
    def __init__(self, registry):

        self.registry = registry
        self.service = {}
        self.service['name'] = "Node-Mock-Backend"
        self.service['type'] = "x-mock"
        self.service['pid'] = 999
        self.service['href'] = "http://localhost/bobbins"
        self.service['proxyHref'] = "localhost/bobbins"

        mocks = {}

        mocks['mockSource'] = {
            "max_api_version": "v1.1",
            "description": "Capture Card Source Video",
            "format": "urn:x-nmos:format:video",
            "tags": {
                "SourceDeviceType": [
                    "HD Camera"
                ],
                "host": [
                    "host1"
                ],
                "Location": [
                    "Location 1"
                ]
            },
            "caps": {},
            "version": "1441722516:851371645",
            "parents": [],
            "label": "CaptureCardSourceVideo",
            "id": "c23c6a65-8e91-4f6c-a484-046363dbca29",
            "device_id": "65fa8c20-890e-4b86-87b2-cfd9df91b7f8",
            "clock_name": "clk1"
        }

        mocks['mockFlow'] = {
            "description": "Test Card",
            "tags": {},
            "format": "urn:x-nmos:format:video",
            "label": "Test Card",
            "version": "1441704616:587121295",
            "parents": [],
            "source_id": "02c46999-d532-4c52-905f-2e368a2af6cb",
            "device_id": "9126cc2f-4c26-4c9b-a6cd-93c4381c9be5",
            "id": "5fbec3b1-1b0f-417d-9059-8b94a47197ed",
            "media_type": "video/raw",
            "frame_width": 1920,
            "frame_height": 1080,
            "interlace_mode": "interlaced_tff",
            "colorspace": "BT709",
            "components": [
                {
                    "name": "Y",
                    "width": 1920,
                    "height": 1080,
                    "bit_depth": 10
                },
                {
                    "name": "Cb",
                    "width": 960,
                    "height": 1080,
                    "bit_depth": 10
                },
                {
                    "name": "Cr",
                    "width": 960,
                    "height": 1080,
                    "bit_depth": 10
                }
            ]
        }

        mocks['mockDeviceA'] = {
            "receivers": [],
            "label": "pipeline 3 default device",
            "description": "pipeline 3 default device",
            "tags": {},
            "version": "1441704616:592733242",
            "id": "9126cc2f-4c26-4c9b-a6cd-93c4381c9be5",
            "type": "urn:x-nmos:device:pipeline",
            "senders": [
                "d7aa5a30-681d-4e72-92fb-f0ba0f6f4c3e"
            ],
            "node_id": "3b8be755-08ff-452b-b217-c9151eb21193",
            "controls": [
                {
                    "type": "urn:x-manufacturer:control:generic",
                    "href": "wss://154.67.63.2:4535"
                }
            ]
        }

        mocks['mockDeviceB'] = {
            "receivers": [
                "1eb53d65-ac83-441c-86f6-9b27df30ef0c"
            ],
            "label": "pipeline 2 default device",
            "description": "pipeline 2 default device",
            "tags": {},
            "version": "1441704514:993221361",
            "id": "05017e08-b329-45f9-a566-a3f99cc11e4d",
            "type": "urn:x-nmos:device:pipeline",
            "senders": [],
            "node_id": "3b8be755-08ff-452b-b217-c9151eb21193",
            "controls": []
        }

        mocks['mockSender'] = {
            "description": "Test Card",
            "label": "Test Card",
            "version": "1441704616:890020555",
            "manifest_href": """
http://172.29.80.65:12345/x-nmos/node/v1.1/self/pipelinemanager/run/pipeline/3/pipel/ipp_rtptxfefa/misc/sdp/stream.sdp
""",
            "flow_id": "5fbec3b1-1b0f-417d-9059-8b94a47197ed",
            "id": "d7aa5a30-681d-4e72-92fb-f0ba0f6f4c3e",
            "transport": "urn:x-nmos:transport:rtp.mcast",
            "device_id": "9126cc2f-4c26-4c9b-a6cd-93c4381c9be5",
            "tags": {}
        }

        mocks['mockReceiver'] = {
            "description": "RTP receiver 1",
            "tags": {
                "Location": [
                    "Location 1"
                ]
            },
            "format": "urn:x-nmos:format:video",
            "caps": {
                "media_types": [
                    "video/raw"
                ]
            },
            "device_id": "0d0cb97e-b96a-4a39-887f-d491492d9081",
            "version": "1441895693:480000000",
            "label": "Viewer 1",
            "id": "1eb53d65-ac83-441c-86f6-9b27df30ef0c",
            "transport": "urn:x-nmos:transport:rtp",
            "subscription": {
                "sender_id": "d7aa5a30-681d-4e72-92fb-f0ba0f6f4c3e"
            }
        }

        self.dataGreenlet = gevent.spawn(self._runMockService, self.registry, self.service, mocks)
        self.heartbeatGreenlet = gevent.spawn(self._runHeartbeat, self.registry, self.service)

    def _runMockService(self, registry, service, mocks):
        registry.register_service(
            service['name'], service['type'], service['pid'], service['href'], service['proxyHref']
        )
        while(True):
            registry.register_resource(
                service['name'], service['pid'], "source", mocks['mockSource']['id'], mocks['mockSource']
            )
            gevent.sleep(10)
            registry.unregister_resource(
                service['name'], service['pid'], "source", mocks['mockSource']['id']
            )
            gevent.sleep(4)
        registry.unregister_service(service['name'], service['pid'])

    def _runHeartbeat(self, registry, service):
        while(True):
            gevent.sleep(11)
            registry.heartbeat_service(service['name'], service['pid'])



# Copyright 2019 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function, absolute_import
from gevent import monkey
monkey.patch_all()

import gevent # noqa E402
import time # noqa E402
import signal # noqa E402
import os  # noqa E402
import sys # noqa E402
import json # noqa E402
import socket # noqa E402

from six import itervalues # noqa E402
from socket import gethostname, getfqdn # noqa E402
from os import getpid, environ # noqa E402
from subprocess import check_output # noqa E402
# Handle if systemd is installed instead of newer cysystemd
try:
    from cysystemd import daemon # noqa E402
    SYSTEMD_READY = daemon.Notification.READY
except ImportError:
    from systemd import daemon # noqa E402
    SYSTEMD_READY = "READY=1"

from nmoscommon.mdns import MDNSEngine # noqa E402
from nmoscommon.logger import Logger # noqa E402
from nmoscommon import ptptime # noqa E402
from nmoscommon.httpserver import HttpServer # noqa E402
from nmoscommon.utils import get_node_id, translate_api_version, getLocalIP # noqa E402
from nmoscommon.nmoscommonconfig import config as _config # noqa E402

from .api import NODE_APIVERSIONS, NODE_REGVERSION, FacadeAPI # noqa E402
from .registry import FacadeRegistry, FacadeRegistryCleaner # noqa E402
from .aggregator import Aggregator, MDNSUpdater # noqa E402
from .authclient import AuthRegistry # noqa E402
from .serviceinterface import FacadeInterface # noqa E402

NS = 'urn:x-bbcrd:ips:ns:0.1'
PORT = 12345
HOSTNAME = gethostname().split(".", 1)[0]
FQDN = getfqdn()

# HTTPS under test only at present
# enabled = Use HTTPS only in all URLs and mDNS adverts
# disabled = Use HTTP only in all URLs and mDNS adverts
# mixed = Use HTTP in all URLs, but additionally advertise an HTTPS endpoint for discovery of this API only
HTTPS_MODE = _config.get('https_mode', 'disabled')
ENABLE_P2P = _config.get('node_p2p_enable', True)
OAUTH_MODE = _config.get('oauth_mode', False)

# BYPASS AUTHLIB SECURITY CHECK DUE TO REVERSE PROXY
environ["AUTHLIB_INSECURE_TRANSPORT"] = "1"


def updateHost():
    if _config.get('node_hostname') is not None:
        return _config.get('node_hostname')
    elif _config.get('prefer_hostnames', False) is True:
        return FQDN
    elif _config.get('prefer_ipv6', False) is False:
        return getLocalIP()
    else:
        return "[" + getLocalIP(None, socket.AF_INET6) + "]"


HOST = updateHost()
DNS_SD_HTTP_PORT = 80
DNS_SD_HTTPS_PORT = 443
DNS_SD_NAME = 'node_' + str(HOSTNAME) + "_" + str(getpid())
DNS_SD_TYPE = '_nmos-node._tcp'


class NodeFacadeService:
    def __init__(self, interactive=False):
        self.logger = Logger("facade", None)
        if HOST == "":
            self.logger.writeFatal("Unable to start facade due to lack of connectivity")
            sys.exit(1)
        self.running = False
        self.httpServer = None
        self.interface = None
        self.interactive = interactive
        self.registry = None
        self.registry_cleaner = None
        self.node_id = None
        self.mdns = MDNSEngine()
        self.mappings = {
            "device": "ver_dvc",
            "flow": "ver_flw",
            "source": "ver_src",
            "sender": "ver_snd",
            "receiver": "ver_rcv",
            "self": "ver_slf"
        }
        self.mdns_updater = None
        self.auth_registry = AuthRegistry()

        if HTTPS_MODE == "enabled":
            self.dns_sd_port = DNS_SD_HTTPS_PORT
            self.protocol = "https"
        else:
            self.dns_sd_port = DNS_SD_HTTP_PORT
            self.protocol = "http"
        if ENABLE_P2P:
            self.mdns_updater = MDNSUpdater(
                self.mdns, DNS_SD_TYPE, DNS_SD_NAME, self.mappings, self.dns_sd_port, self.logger,
                txt_recs=self._mdns_txt(NODE_APIVERSIONS, self.protocol, OAUTH_MODE)
            )

        self.aggregator = Aggregator(self.logger, self.mdns_updater, self.auth_registry)

    def _mdns_txt(self, versions, protocol, oauth_mode):
        return {
            "api_ver": ",".join(versions),
            "api_proto": protocol,
            "api_auth": str(oauth_mode).lower()
        }

    def sig_handler(self):
        print('Pressed ctrl+c')
        self.stop()

    def sig_hup_handler(self):
        if getLocalIP() != "":
            global HOST
            HOST = updateHost()
            self.registry.modify_node(
                href=self.generate_href(),
                host=HOST,
                api={"versions": NODE_APIVERSIONS, "endpoints": self.generate_endpoints()},
                interfaces=self.list_interfaces()
            )

    def generate_endpoints(self):
        endpoints = []
        endpoints.append({
            "host": HOST,
            "port": self.dns_sd_port,  # Everything should go via apache proxy
            "protocol": self.protocol,
            "authorization": OAUTH_MODE
        })
        return endpoints

    def generate_href(self):
        if HTTPS_MODE == "enabled":
            return "https://{}/".format(HOST)
        else:
            return "http://{}/".format(HOST)

    def list_interfaces(self):
        interfaces = {}
        # Initially populate interfaces from known-good location
        net_path = "/sys/class/net/"
        if os.path.exists(net_path):
            for interface_name in os.listdir(net_path):
                address_path = net_path + interface_name + "/address"
                if os.path.exists(address_path) and interface_name != 'lo':
                    with open(address_path, 'r') as address_file:
                        address = address_file.readline().strip('\n')
                        if address:
                            interfaces[interface_name] = {
                                "name": interface_name,
                                "chassis_id": None,
                                "port_id": address.lower().replace(":", "-")
                            }

        # Attempt to source proper LLDP data for interfaces
        if os.path.exists("/usr/sbin/lldpcli"):
            try:
                chassis_data = json.loads(check_output(["/usr/sbin/lldpcli", "show", "chassis", "-f", "json"]))
                chassis_id = chassis_data["local-chassis"]['chassis'].values()[0]["id"]["value"]
                if chassis_data["local-chassis"]['chassis'].values()[0]["id"]["type"] == "mac":
                    chassis_id = chassis_id.lower().replace(":", "-")
                interface_data = json.loads(check_output(["/usr/sbin/lldpcli", "show", "statistics", "-f", "json"]))
                if isinstance(interface_data["lldp"]["interface"], dict):
                    for interface_name in interface_data["lldp"]["interface"].keys():
                        if interface_name in interfaces:
                            # Only correct the Chassis ID. Port ID MUST be a MAC address
                            interfaces[interface_name]["chassis_id"] = chassis_id
                else:
                    for interface_block in interface_data["lldp"]["interface"]:
                        interface_name = interface_block.keys()[0]
                        if interface_name in interfaces:
                            # Only correct the Chassis ID. Port ID MUST be a MAC address
                            interfaces[interface_name]["chassis_id"] = chassis_id
            except Exception:
                pass

        return list(itervalues(interfaces))

    def start(self):
        if self.running:
            gevent.signal(signal.SIGINT, self.sig_handler)
            gevent.signal(signal.SIGTERM, self.sig_handler)
            gevent.signal(signal.SIGHUP, self.sig_hup_handler)

        self.mdns.start()
        self.node_id = get_node_id()
        node_version = str(ptptime.ptp_detail()[0]) + ":" + str(ptptime.ptp_detail()[1])
        node_data = {
            "id": self.node_id,
            "label": _config.get('node_label', FQDN),
            "description": _config.get('node_description', "Node on {}".format(FQDN)),
            "tags": _config.get('node_tags', {}),
            "href": self.generate_href(),
            "host": HOST,
            "services": [],
            "hostname": HOSTNAME,
            "caps": {},
            "version": node_version,
            "api": {
                "versions": NODE_APIVERSIONS,
                "endpoints": self.generate_endpoints(),
            },
            "clocks": [],
            "interfaces": self.list_interfaces()
        }
        self.registry = FacadeRegistry(
            self.mappings.keys(),
            self.aggregator,
            self.mdns_updater,
            self.node_id,
            node_data,
            self.logger
        )
        self.registry_cleaner = FacadeRegistryCleaner(self.registry)
        self.registry_cleaner.start()
        self.httpServer = HttpServer(
            FacadeAPI, PORT, '0.0.0.0', api_args=[self.registry, self.auth_registry])
        self.httpServer.start()
        while not self.httpServer.started.is_set():
            self.logger.writeInfo('Waiting for httpserver to start...')
            self.httpServer.started.wait()

        if self.httpServer.failed is not None:
            raise self.httpServer.failed

        self.logger.writeInfo("Running on port: {}".format(self.httpServer.port))

        try:
            self.logger.writeInfo("Registering as {}...".format(self.node_id))
            self.aggregator.register('node', self.node_id, **translate_api_version(node_data, "node", NODE_REGVERSION))
        except Exception as e:
            self.logger.writeWarning("Could not register: {}".format(e.__repr__()))

        self.interface = FacadeInterface(self.registry, self.logger)
        self.interface.start()

    def run(self):
        self.running = True
        pidfile = "/tmp/ips-nodefacade.pid"
        with open(pidfile, 'w') as f:
            f.write(str(getpid()))
        self.start()
        daemon.notify(SYSTEMD_READY)
        while self.running:
            self.registry.update_ptp()
            time.sleep(1)
        os.unlink(pidfile)

    def _cleanup(self):
        try:
            self.logger.writeDebug("cleanup: unregister facade " + self.node_id)
            self.aggregator.unregister('node', self.node_id)
        except Exception as e:
            self.logger.writeWarning("Could not unregister: {}".format(e))

        if self.mdns:
            try:
                self.mdns.stop()
            except Exception as e:
                self.logger.writeWarning("Could not stop mdns: {}".format(e))

        self.registry_cleaner.stop()
        self.interface.stop()
        self.httpServer.stop()
        self.aggregator.stop()
        self.mdns_updater.stop()
        self.logger.writeInfo("Stopped main()")

    def stop(self):
        self._cleanup()
        self.running = False


if __name__ == '__main__':
    Service = NodeFacadeService()
    Service.run()
